# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Iterable
from dataclasses import dataclass, field
from typing import Union

from .gene import Gene
from .mutation import Mutation
from .result_with_gene import ResultWithGene


@dataclass(eq=False, repr=False, frozen=True, init=False)
class Allele(ResultWithGene):
    """
    Represents an MHC allele with gene and allele field information.

    An Allele specifies a unique protein amino acid sequence using notation
    like "HLA-A*02:01" or more generally: Species-Gene*Group:ProteinID

    Alleles can have annotations (e.g., "N" for null, "S" for secreted)
    and mutations (e.g., "N80I").

    Parameters
    ----------
    gene : Gene
        The gene this allele belongs to.
    allele_fields : Iterable[str]
        The allele field identifiers (e.g., ("02", "01") for A*02:01).
    annotations : Iterable[str], optional
        Functional annotations like "N" (null), "S" (secreted), etc.
    mutations : Iterable[Mutation], optional
        Any mutations from the reference sequence.
    raw_string : str, optional
        The original unparsed string.

    Attributes
    ----------
    gene : Gene
        The gene this allele belongs to.
    allele_fields : tuple of str
        The allele field identifiers.
    annotations : tuple of str
        Functional annotations.
    mutations : tuple of Mutation
        Any mutations from the reference sequence.

    Examples
    --------
    >>> from mhcgnomes import parse
    >>> allele = parse("HLA-A*02:01")
    >>> allele.gene_name
    'A'
    >>> allele.allele_fields
    ('02', '01')
    >>> allele.to_string()
    'HLA-A*02:01'
    """

    allele_fields: tuple[str, ...] = field(default_factory=tuple)
    annotations: tuple[str, ...] = field(default_factory=tuple)
    mutations: tuple[Mutation, ...] = field(default_factory=tuple)

    def __init__(
        self,
        gene: Gene,
        allele_fields: Iterable[str],
        annotations: Iterable[str] = (),
        mutations: Iterable[Mutation] = (),
        raw_string=None,
    ):
        if gene.is_mutant:
            gene_mutations = tuple(gene.mutations)
            gene = gene.copy_without_mutations()
        else:
            gene_mutations = ()
        ResultWithGene.__init__(self, gene=gene, raw_string=raw_string)
        self._set_field(self, "allele_fields", tuple(allele_fields))
        self._set_field(self, "annotations", tuple(annotations))
        self._set_field(self, "mutations", gene_mutations + tuple(mutations))

    def __hash__(self):
        return hash((self.gene, self.allele_fields, self.annotations, self.mutations))

    def __eq__(self, other):
        if type(other) is not Allele:
            return False
        return (
            self.gene == other.gene
            and self.allele_fields == other.allele_fields
            and self.annotations == other.annotations
            and self.mutations == other.mutations
        )

    @classmethod
    def tuple_field_names(cls):
        return ("gene", "allele_fields", "annotations", "mutations", "raw_string")

    @classmethod
    def str_field_names(cls):
        return (
            "gene",
            "allele_fields",
            "annotations",
            "mutations",
        )

    @classmethod
    def eq_field_names(cls):
        return (
            "gene",
            "allele_fields",
            "annotations",
            "mutations",
        )

    @property
    def num_allele_fields(self):
        return len(self.allele_fields)

    @property
    def name(self):
        return ":".join(self.allele_fields) + "".join(self.annotations)

    def restrict_allele_fields(self, num_fields, drop_annotations=False, drop_mutations=False):
        if num_fields >= self.num_allele_fields:
            return self
        else:
            return self.copy(
                allele_fields=self.allele_fields[:num_fields],
                annotations=() if drop_annotations else self.annotations,
                mutations=() if drop_mutations else self.mutations,
            )

    @classmethod
    def split_allele_fields(cls, allele_fields):
        if isinstance(allele_fields, str):
            allele_fields = allele_fields.split(":")
        allele_fields = [str(s).strip() for s in allele_fields]
        allele_fields = [s for s in allele_fields if len(s) > 0]
        return tuple(allele_fields)

    @classmethod
    def get_with_gene(
        cls,
        gene: Gene,
        allele_fields: Iterable[str],
        annotations: Union[Iterable[str], None] = None,
        mutations: Union[Iterable[Mutation], None] = None,
        raw_string: Union[str, None] = None,
    ):
        """
        Create an Allele from a Gene object and allele fields.

        Parameters
        ----------
        gene : Gene
            The gene this allele belongs to.
        allele_fields : Iterable[str]
            The allele field identifiers.
        annotations : Iterable[str], optional
            Functional annotations.
        mutations : Iterable[Mutation], optional
            Mutations from reference sequence.
        raw_string : str, optional
            Original unparsed string.

        Returns
        -------
        Allele or None
            The constructed Allele, or None if gene is None or no allele
            fields provided.
        """
        if gene is None:
            return None

        allele_fields = cls.split_allele_fields(allele_fields)

        if len(allele_fields) == 0:
            return None

        if annotations is None:
            annotations = ()

        if mutations is None:
            mutations = ()

        return Allele(
            gene=gene,
            allele_fields=allele_fields,
            annotations=annotations,
            mutations=mutations,
            raw_string=raw_string,
        )

    @classmethod
    def get(
        cls,
        species_prefix,
        gene_name,
        *allele_fields,
        annotation=None,
        mutations=None,
        raw_string=None,
    ):
        """
        Create an Allele from species prefix, gene name, and allele fields.

        This is the primary factory method for creating Allele objects
        programmatically.

        Parameters
        ----------
        species_prefix : str
            Species prefix (e.g., "HLA" for human).
        gene_name : str
            Gene name (e.g., "A", "B", "DRB1").
        *allele_fields : str
            Variable number of allele field arguments, or a single
            colon-separated string.
        annotation : str, optional
            Single functional annotation.
        mutations : str or list, optional
            Mutations as string or list of Mutation objects.
        raw_string : str, optional
            Original unparsed string.

        Returns
        -------
        Allele or None
            The constructed Allele, or None if species/gene not found.

        Examples
        --------
        >>> allele = Allele.get("HLA", "A", "02", "01")
        >>> allele is not None
        True
        >>> allele.allele_fields
        ('02', '01')
        >>> Allele.get("HLA", "A", "02:01").allele_fields  # Also valid
        ('02', '01')
        """
        gene = Gene.get(species_prefix, gene_name)
        if not gene:
            return None

        if len(allele_fields) == 1:
            allele_fields = cls.split_allele_fields(allele_fields[0])

        if len(allele_fields) == 0:
            return None

        if annotation:
            annotations = [annotation]
        else:
            annotations = []

        if mutations is None:
            mutations = []

        elif type(mutations) is str:
            mutations = mutations.split()

        parsed_mutations = []
        for mut in mutations:
            if type(mut) is str:
                mut = Mutation.parse(mut, raise_on_error=False)
            if mut is None:
                return None
            if type(mut) is not Mutation:
                return None
            parsed_mutations.append(mut)
        return Allele.get_with_gene(
            gene=gene,
            allele_fields=allele_fields,
            annotations=annotations,
            mutations=parsed_mutations,
            raw_string=raw_string,
        )

    def to_string(
        self, include_species=True, use_old_species_prefix=False, include_annotations=True
    ):
        """
        Return allele strings like "HLA-A*02:01"
        """
        gene_str = self.gene.to_string(
            include_species=include_species, use_old_species_prefix=use_old_species_prefix
        )
        result = "{}*{}".format(gene_str, ":".join(self.allele_fields))
        if include_annotations:
            for annot in self.annotations:
                result += annot
        if self.is_mutant:
            result += f" {self.mutation_string()} mutant"
        return result

    def compact_string(
        self, include_species=False, use_old_species_prefix=False, include_annotations=False
    ):
        """
        Compact representation of a Allele, omits the "*" and ":"
        in allele names
            Normalized: HLA-A*02:01
            Compact: HLA-A0201
        """
        result = "{}{}{}".format(
            self.gene.compact_string(
                include_species=include_species, use_old_species_prefix=use_old_species_prefix
            ),
            "*" if self.gene_name[-1].isdigit() else "",
            "".join(self.allele_fields),
        )
        if self.is_mutant:
            result += f" {self.mutation_string()} mutant"
        return result

    def copy_with_extra_mutations(self, mutations):
        new_mutations = tuple(self.mutations) + tuple(mutations)
        return self.copy(mutations=new_mutations)

    @property
    def is_mutant(self):
        return len(self.mutations) > 0

    def mutation_string(self):
        return " ".join([mut.to_string() for mut in self.mutations])

    def to_record(self):
        """
        Returns dictionary with all fields of this allele,
        as well as its representations as a gene, allele group,
        and four digit allele.
        """
        d = self.gene.to_record()
        d["allele"] = self.to_string()
        d["annotations"] = self.annotations
        d["mutations"] = self.mutation_string()
        d["is_mutant"] = self.is_mutant
        return d

    def _check_annotation(self, annotation):
        """
        Check if given annotation is in the annotation list.
        """
        return annotation in self.annotations

    @property
    def annotation_null(self):
        return self._check_annotation("N")

    @property
    def annotation_cystosolic(self):
        return self._check_annotation("C")

    @property
    def annotation_secreted(self):
        return self._check_annotation("S")

    @property
    def annotation_questionable(self):
        return self._check_annotation("Q")

    @property
    def annotation_low_expression(self):
        return self._check_annotation("L")

    @property
    def annotation_aberrant_expression(self):
        return self._check_annotation("A")

    @property
    def annotation_group(self):
        # designates a group of genomic sequence alleles
        # with identical peptide binding region
        return self._check_annotation("G")

    @property
    def annotation_pseudogene(self):
        # designates a group of genomic sequence alleles
        # with identical peptide binding region
        return self._check_annotation("Ps")

    @property
    def annotation_splice_variant(self):
        # designates a group of genomic sequence alleles
        # with identical peptide binding region
        return self._check_annotation("Sp")



"""
Tests for taxonomic gene inheritance and cross-taxon leak prevention.

Species inherit genes from their parent node (e.g., salmonids from
Salmonidae sp.). Root-level transport and DM genes come from
Gnathostomata sp. through the tree.
Taxon-specific genes must NOT leak across boundaries.
"""

import pytest

from mhcgnomes import Gene, MhcClass, Species, parse
from mhcgnomes.species import raw_species_dict

from .common import eq_

GNATHOSTOMATA = "Gnathostomata sp."
ROOTED_SPECIES = sorted(latin for latin in raw_species_dict if latin != GNATHOSTOMATA)
INTENTIONALLY_UNROOTED = set()


def has_ancestor(species, ancestor_name):
    current = species
    while current is not None:
        if current.name == ancestor_name:
            return True
        current = current.parent_species
    return False


# ---------------------------------------------------------------------------
# Salmonid inheritance
# ---------------------------------------------------------------------------


class TestSalmonidInheritance:
    """Salmonid species inherit UBA, DAA, DAB from Salmonidae sp."""

    def test_salmonid_parent_has_genes(self):
        sp = Species.get("Salmonidae sp.")
        assert sp is not None
        assert sp.find_matching_gene_name("UBA") is not None
        assert sp.find_matching_gene_name("DAA") is not None
        assert sp.find_matching_gene_name("DAB") is not None

    def test_rainbow_trout_inherits_uba(self):
        eq_(parse("Onmy-UBA", raise_on_error=True).species.prefix, "Onmy")

    def test_atlantic_salmon_inherits_daa(self):
        eq_(parse("Sasa-DAA", raise_on_error=True).species.prefix, "Sasa")

    def test_brown_trout_inherits_uba(self):
        eq_(parse("Satr-UBA", raise_on_error=True).species.prefix, "Satr")

    def test_brown_trout_inherits_daa(self):
        """Brown trout previously had DAA only as a species-specific gene.
        Now it inherits from the salmonid parent."""
        eq_(parse("Satr-DAA", raise_on_error=True).species.prefix, "Satr")

    def test_arctic_char_inherits_plus_own_genes(self):
        """Arctic char inherits salmonid genes AND has UEA, UGA of its own."""
        sp = Species.get("Saal")
        assert sp.find_matching_gene_name("UBA") is not None  # inherited
        assert sp.find_matching_gene_name("UEA") is not None  # own
        assert sp.find_matching_gene_name("UGA") is not None  # own
        assert sp.find_matching_gene_name("DAB") is not None  # inherited


# ---------------------------------------------------------------------------
# Common genes (truly universal)
# ---------------------------------------------------------------------------


class TestGnathostomataRootInheritance:
    """DMA, DMB, TAP1, TAP2, TAPBP, B2M inherited from Gnathostomata root."""

    @pytest.mark.parametrize(
        "latin",
        [
            "Chelonia mydas",  # turtle → Testudines → Gnathostomata
            "Struthio camelus",  # ratite → Gnathostomata (no intermediate)
            "Crocodylus porosus",  # croc → Crocodylia → Gnathostomata
            "Sphenodon punctatus",  # tuatara → Gnathostomata (no intermediate)
        ],
    )
    def test_root_genes_inherited_via_tree(self, latin):
        sp = Species.get(latin)
        assert sp is not None
        for gene in ["DMA", "TAP1"]:
            assert sp.find_matching_gene_name(gene) is not None, (
                f"{latin} should inherit {gene} from Gnathostomata root"
            )

    def test_root_gene_with_default_species(self):
        result = parse("DMA", default_species="Chelonia mydas", raise_on_error=True)
        assert result is not None
        eq_(result.species.name, "Chelonia mydas")


# ---------------------------------------------------------------------------
# Cross-taxon leak prevention
# ---------------------------------------------------------------------------


class TestCrossTaxonLeakPrevention:
    """Taxon-specific genes must NOT be accepted by species from other taxa."""

    def test_human_does_not_get_fish_uaa(self):
        sp = Species.get("HLA")
        assert sp.find_matching_gene_name("UAA") is None

    def test_human_does_not_get_fish_uba(self):
        sp = Species.get("HLA")
        assert sp.find_matching_gene_name("UBA") is None

    def test_human_does_not_get_fish_dab(self):
        sp = Species.get("HLA")
        assert sp.find_matching_gene_name("DAB") is None

    def test_zebrafish_does_not_get_mammal_drb(self):
        sp = Species.get("Dare")
        assert sp.find_matching_gene_name("DRB") is None

    def test_zebrafish_does_not_get_mammal_drb1(self):
        sp = Species.get("Dare")
        assert sp.find_matching_gene_name("DRB1") is None

    def test_chicken_inherits_uaa_from_aves(self):
        """Chicken now inherits UAA from Aves (non-mammalian class I gene)."""
        sp = Species.get("Gaga")
        assert sp.find_matching_gene_name("UAA") is not None

    def test_chicken_inherits_drb_from_aves(self):
        """Chicken now inherits DRB from Aves clade node."""
        sp = Species.get("Gaga")
        assert sp.find_matching_gene_name("DRB") is not None

    def test_frog_does_not_get_drb(self):
        """Amphibia has DA/DB/DC but not DR — frogs should not have DRB."""
        sp = Species.get("Xela")
        assert sp.find_matching_gene_name("DRB") is None

    def test_parse_hla_uaa_fails(self):
        """HLA-UAA should not parse — UAA is a fish gene, not human."""
        assert parse("HLA-UAA", raise_on_error=False) is None

    def test_parse_satr_dqa_fails(self):
        """Satr-DQA should not parse — DQA is a mammalian gene, not fish."""
        assert parse("Satr-DQA", raise_on_error=False) is None


# ---------------------------------------------------------------------------
# default_species priority
# ---------------------------------------------------------------------------


class TestGalliformInheritance:
    """Galliform species inherit BF, BLB, DMA, DMB1/2, TAP1/2 from parent."""

    def test_galliform_parent_has_genes(self):
        sp = Species.get("Galliformes sp.")
        assert sp is not None
        assert sp.find_matching_gene_name("BF") is not None
        assert sp.find_matching_gene_name("BLB") is not None
        assert sp.find_matching_gene_name("DMA") is not None
        assert sp.find_matching_gene_name("TAP1") is not None

    def test_turkey_inherits_bf(self):
        """Turkey inherits BF from Galliformes sp. parent."""
        result = parse("Mega-BF", raise_on_error=True)
        eq_(result.species.name, "Meleagris gallopavo")

    def test_peafowl_inherits_blb(self):
        result = parse("Pacr-BLB", raise_on_error=True)
        eq_(result.species.name, "Pavo cristatus")

    def test_chicken_still_has_specific_genes(self):
        """Chicken should still have BF1, BF2, YF1, YF2 on top of inherited BF."""
        sp = Species.get("Gaga")
        assert sp.find_matching_gene_name("BF1") is not None
        assert sp.find_matching_gene_name("BF2") is not None
        assert sp.find_matching_gene_name("YF1") is not None

    def test_quail_has_inherited_plus_own(self):
        """Quail inherits BF/BLB from Galliformes and has own class I/II loci."""
        sp = Species.get("Coja")
        assert sp.find_matching_gene_name("BF") is not None  # inherited
        assert sp.find_matching_gene_name("DAB1") is not None  # own

    def test_galliform_inherits_aves_genes(self):
        """Galliforms now inherit non-mammalian U-lineage and DR genes from Aves."""
        sp = Species.get("Gaga")
        assert sp.find_matching_gene_name("UAA") is not None
        assert sp.find_matching_gene_name("UBA") is not None
        assert sp.find_matching_gene_name("DRB") is not None

    def test_galliform_inherits_drb1_from_aves(self):
        """DRB1 is now on Aves — galliforms inherit it."""
        sp = Species.get("Mega")
        assert sp.find_matching_gene_name("DRB1") is not None

    def test_galliform_does_not_get_mammal_specific_genes(self):
        """Galliforms should NOT have mammal-specific DQ genes."""
        sp = Species.get("Mega")
        assert sp.find_matching_gene_name("DQA1") is None


class TestCrocodyliaInheritance:
    """Crocodilian species inherit UA and DB01-DB08 from Crocodylia sp. parent."""

    def test_crocodylia_parent_has_genes(self):
        sp = Species.get("Crocodylia sp.")
        assert sp is not None
        assert sp.find_matching_gene_name("UA") is not None
        assert sp.find_matching_gene_name("DB01") is not None
        assert sp.find_matching_gene_name("DB08") is not None

    def test_alligator_inherits_db_genes(self):
        """American alligator inherits DB01-DB08 from Crocodylia sp."""
        result = parse("Almi-DB01", raise_on_error=True)
        eq_(result.species.name, "Alligator mississippiensis")

    def test_gharial_inherits_ua(self):
        result = parse("GaviaGange-UA", raise_on_error=True)
        eq_(result.species.name, "Gavialis gangeticus")

    def test_crpo_has_inherited_plus_own(self):
        """Crpo inherits UA/DB01-08 and has own UB, UC, DAA, DAB1, DAB2."""
        sp = Species.get("Crpo")
        assert sp.find_matching_gene_name("UA") is not None  # inherited
        assert sp.find_matching_gene_name("DB01") is not None  # inherited
        assert sp.find_matching_gene_name("UB") is not None  # own
        assert sp.find_matching_gene_name("DAA") is not None  # own

    def test_croc_does_not_get_mammal_genes(self):
        sp = Species.get("Crpo")
        assert sp.find_matching_gene_name("DRB") is None
        # DRB1 is now on Reptilia
        assert sp.find_matching_gene_name("DRB1") is not None
        assert sp.find_matching_gene_name("DQA1") is None

    def test_croc_does_not_get_galliform_genes(self):
        sp = Species.get("Crpo")
        assert sp.find_matching_gene_name("BF") is None
        assert sp.find_matching_gene_name("BF1") is None
        assert sp.find_matching_gene_name("BLB") is None


class TestDMBInheritance:
    """DMB comes from Gnathostomata root; DMB1/DMB2 from Galliformes or species."""

    def test_chicken_has_dmb_from_root_and_dmb1_from_galliformes(self):
        """Chicken inherits DMB from Gnathostomata AND DMB1/DMB2 from Galliformes."""
        sp = Species.get("Gaga")
        assert sp.find_matching_gene_name("DMB") is not None
        assert sp.find_matching_gene_name("DMB1") is not None
        assert sp.find_matching_gene_name("DMB2") is not None

    def test_turtle_gets_dmb_from_root(self):
        """Turtle inherits DMB from Gnathostomata via Testudines."""
        sp = Species.get("Chelonia mydas")
        assert sp.find_matching_gene_name("DMB") is not None
        assert sp.find_matching_gene_name("DMA") is not None
        assert sp.find_matching_gene_name("TAP1") is not None


class TestDefaultSpeciesPriority:
    """When default_species is provided, it takes priority over gene-inferred species."""

    def test_dma_with_turtle_default(self):
        """Turtle inherits DMA from Gnathostomata root."""
        result = parse("DMA", default_species="Chelonia mydas", raise_on_error=True)
        eq_(result.species.name, "Chelonia mydas")

    def test_tap1_with_tuatara_default(self):
        """Tuatara inherits TAP1 from Gnathostomata root."""
        result = parse("TAP1", default_species="Sphenodon punctatus", raise_on_error=True)
        eq_(result.species.name, "Sphenodon punctatus")

    def test_dmb_with_ostrich_default(self):
        """Ostrich inherits DMB from Gnathostomata root."""
        result = parse("DMB", default_species="Struthio camelus", raise_on_error=True)
        eq_(result.species.name, "Struthio camelus")


class TestRootTransportGenesAndAliases:
    """All rooted species inherit the jawed-vertebrate transport/DM core."""

    @pytest.mark.parametrize("latin", ROOTED_SPECIES)
    def test_all_species_descend_from_gnathostomata(self, latin):
        sp = Species.get(latin)
        assert sp is not None
        assert has_ancestor(sp, GNATHOSTOMATA), latin

    def test_only_intentionally_unrooted_species_lack_root_ancestor(self):
        unrooted = {
            latin
            for latin in raw_species_dict
            if latin != GNATHOSTOMATA and not has_ancestor(Species.get(latin), GNATHOSTOMATA)
        }
        eq_(unrooted, INTENTIONALLY_UNROOTED)

    @pytest.mark.parametrize("latin", ROOTED_SPECIES)
    def test_all_species_inherit_root_transport_and_dm_genes(self, latin):
        sp = Species.get(latin)
        assert sp is not None
        for gene_name in ["DMA", "DMB", "TAP1", "TAP2", "TAP-L", "TAPBP", "B2M"]:
            assert sp.find_matching_gene_name(gene_name) is not None, (latin, gene_name)
        assert sp.find_matching_class2_locus_name("DM") is not None, latin

    @pytest.mark.parametrize(
        "alias, expected",
        [
            ("ABCB2", "TAP1"),
            ("RING4", "TAP1"),
            ("PSF1", "TAP1"),
            ("HAM1", "TAP1"),
            ("ABCB3", "TAP2"),
            ("RING11", "TAP2"),
            ("PSF2", "TAP2"),
            ("HAM2", "TAP2"),
            ("ABCB9", "TAP-L"),
            ("TAPL", "TAP-L"),
        ],
    )
    def test_transport_aliases_parse_to_canonical_root_genes(self, alias, expected):
        result = parse(alias, default_species="Struthio camelus", raise_on_error=True)
        eq_(result.species.name, "Struthio camelus")
        eq_(result.name, expected)

    @pytest.mark.parametrize("raw", ["TAP1", "TAP2", "TAP-L", "TAPBP", "B2M"])
    def test_root_other_genes_parse_with_default_species(self, raw):
        result = parse(raw, default_species="Struthio camelus", raise_on_error=True)
        eq_(result.species.name, "Struthio camelus")
        eq_(result.name, raw)

    def test_dm_locus_parses_with_default_species(self):
        result = parse("DM", default_species="Struthio camelus", raise_on_error=True)
        eq_(result.species.name, "Struthio camelus")
        eq_(result.name, "DM")


class TestMhciibScoping:
    """MHCIIB is only accepted for species that actually use that alias."""

    def test_mhciib_parses_as_mhc_class_for_any_species(self):
        """MHCIIB is now a generic class II beta label, works for any species."""
        result = parse("MHCIIB", default_species="Struthio camelus", raise_on_error=True)
        assert isinstance(result, MhcClass)
        eq_(result.chain, "beta")
        eq_(result.species.name, "Struthio camelus")

    def test_mhciib_with_strict_species(self):
        result = parse("MHCIIB", species="Struthio camelus", raise_on_error=True)
        assert isinstance(result, MhcClass)
        eq_(result.species.name, "Struthio camelus")

    def test_mhciib_parses_as_mhc_class_not_gene(self):
        """MHCIIB is now a class II beta MhcClass, not a Gene(Tyal, DAB)."""
        result = parse("MHCIIB", default_species="Tyto alba", raise_on_error=True)
        assert isinstance(result, MhcClass)
        eq_(result.mhc_class, "II")
        eq_(result.chain, "beta")


class TestSpeciesStrictness:
    """species= should reject mismatched species, including Species results."""

    def test_species_argument_rejects_mismatched_species_result(self):
        assert parse("HLA", species="H-2", raise_on_error=False) is None

    def test_species_argument_rejects_mismatched_taxonomic_node(self):
        assert parse("Galliformes", species="Gallus gallus", raise_on_error=False) is None

    def test_species_argument_accepts_matching_taxonomic_node(self):
        result = parse("Galliformes", species="Galliformes sp.", raise_on_error=True)
        eq_(result.name, "Galliformes sp.")


class TestTaxonomicPrefixLeakage:
    """Taxonomic node prefixes must not resolve child-only genes or outputs."""

    @pytest.mark.parametrize("raw", ["Galliformes-BF1", "Crocodylia-UB", "Salmonidae-UEA"])
    def test_taxonomic_prefixes_do_not_resolve_child_specific_genes(self, raw):
        assert parse(raw, raise_on_error=False) is None

    @pytest.mark.parametrize(
        "species_prefix, gene_name, expected_string",
        [
            ("Crpo", "UB", "Crpo-UB"),
            ("Gaga", "BF1", "Gaga-BF1"),
            ("Satr", "UBA", "Satr-UBA"),
        ],
    )
    def test_use_old_species_prefix_does_not_emit_taxonomic_prefix(
        self, species_prefix, gene_name, expected_string
    ):
        gene = Gene.get(species_prefix, gene_name)
        assert gene is not None
        eq_(gene.to_string(use_old_species_prefix=True), expected_string)


# ---------------------------------------------------------------------------
# Intermediate clade node inheritance
# ---------------------------------------------------------------------------


class TestChondrichthyesInheritance:
    """Chondrichthyes sp. provides UA, DA, DB, DC genes to sharks."""

    def test_clade_has_genes(self):
        sp = Species.get("Chondrichthyes sp.")
        assert sp is not None
        assert sp.find_matching_gene_name("UA") is not None
        assert sp.find_matching_gene_name("DAA") is not None
        assert sp.find_matching_gene_name("DAB") is not None
        assert sp.find_matching_gene_name("DBA") is not None
        assert sp.find_matching_gene_name("DCB") is not None

    def test_shark_inherits_ua(self):
        sp = Species.get("Gici")
        assert sp.find_matching_gene_name("UA") is not None

    def test_shark_inherits_dab(self):
        sp = Species.get("Hefr")
        assert sp.find_matching_gene_name("DAB") is not None

    def test_shark_inherits_root_genes(self):
        sp = Species.get("Sqac")
        assert sp.find_matching_gene_name("DMA") is not None
        assert sp.find_matching_gene_name("TAP1") is not None

    def test_shark_does_not_get_mammal_genes(self):
        sp = Species.get("Trsc")
        assert sp.find_matching_gene_name("DRB") is None
        assert sp.find_matching_gene_name("DRB1") is None


class TestActinopterygiiInheritance:
    """Actinopterygii sp. provides fish-wide class I/II genes."""

    def test_clade_has_genes(self):
        sp = Species.get("Actinopterygii sp.")
        assert sp is not None
        for gene in [
            "UAA",
            "UBA",
            "UA",
            "UCA",
            "UDA",
            "DAA",
            "DAB",
            "DBA",
            "DBB",
            "DCA",
            "DCB",
            "DXA",
            "DXB",
        ]:
            assert sp.find_matching_gene_name(gene) is not None, gene

    def test_salmonidae_inherits_from_actinopterygii(self):
        sp = Species.get("Salmonidae sp.")
        assert sp.find_matching_gene_name("UAA") is not None
        assert sp.find_matching_gene_name("DAB") is not None
        assert sp.find_matching_gene_name("DCA") is not None

    def test_zebrafish_inherits_actinopterygii_genes(self):
        sp = Species.get("Dare")
        assert sp.find_matching_gene_name("UAA") is not None
        assert sp.find_matching_gene_name("DAA") is not None
        assert sp.find_matching_gene_name("DXA") is not None

    def test_tropheus_has_dd_genes(self):
        sp = Species.get("Trsp")
        assert sp.find_matching_gene_name("DDA") is not None
        assert sp.find_matching_gene_name("DDB1") is not None
        assert sp.find_matching_gene_name("DDB2") is not None

    def test_coregonus_inherits_actinopterygii(self):
        sp = Species.get("Cosp")
        assert sp.find_matching_gene_name("UAA") is not None

    def test_fish_does_not_get_drb(self):
        sp = Species.get("Dare")
        assert sp.find_matching_gene_name("DRB") is None


class TestAmphibiaInheritance:
    """Amphibia sp. provides class I/II genes to frogs, salamanders, caecilians."""

    def test_clade_has_genes(self):
        sp = Species.get("Amphibia sp.")
        assert sp is not None
        for gene in ["UAA", "UBA", "UA", "UCA", "UDA", "DAA", "DAB", "DBA", "DBB", "DCA", "DCB"]:
            assert sp.find_matching_gene_name(gene) is not None, gene

    def test_xenopus_inherits_class1(self):
        sp = Species.get("Xela")
        assert sp.find_matching_gene_name("UAA") is not None
        assert sp.find_matching_gene_name("UBA") is not None
        assert sp.find_matching_gene_name("UCA") is not None

    def test_xenopus_keeps_own_uea(self):
        sp = Species.get("Xela")
        assert sp.find_matching_gene_name("UEA") is not None

    def test_axolotl_inherits_amphibia(self):
        sp = Species.get("Amme")
        assert sp.find_matching_gene_name("DAA") is not None
        assert sp.find_matching_gene_name("UA") is not None

    def test_amphibian_does_not_get_drb(self):
        sp = Species.get("Xela")
        assert sp.find_matching_gene_name("DRB") is None

    def test_amphibian_does_not_get_dra(self):
        sp = Species.get("Xela")
        assert sp.find_matching_gene_name("DRA") is None


class TestReptiliaInheritance:
    """Reptilia sp. provides class I/II genes to lizards, snakes, tuatara."""

    def test_clade_has_genes(self):
        sp = Species.get("Reptilia sp.")
        assert sp is not None
        for gene in [
            "UAA",
            "UBA",
            "UA",
            "UCA",
            "UDA",
            "DAA",
            "DAB",
            "DBA",
            "DBB",
            "DCA",
            "DCB",
            "DRA",
        ]:
            assert sp.find_matching_gene_name(gene) is not None, gene

    def test_reptilia_does_not_have_drb(self):
        sp = Species.get("Reptilia sp.")
        assert sp.find_matching_gene_name("DRB") is None

    def test_testudines_under_reptilia(self):
        sp = Species.get("Testudines sp.")
        assert has_ancestor(sp, "Reptilia sp.")
        assert sp.find_matching_gene_name("UA") is not None
        assert sp.find_matching_gene_name("DAA") is not None

    def test_crocodylia_under_reptilia(self):
        sp = Species.get("Crocodylia sp.")
        assert has_ancestor(sp, "Reptilia sp.")
        assert sp.find_matching_gene_name("UA") is not None
        assert sp.find_matching_gene_name("DB01") is not None

    def test_tuatara_inherits_reptilia(self):
        sp = Species.get("Sppu")
        assert sp.find_matching_gene_name("UA") is not None
        assert sp.find_matching_gene_name("DAA") is not None
        assert sp.find_matching_gene_name("DRA") is not None

    def test_snake_inherits_reptilia(self):
        sp = Species.get("Nana")
        assert sp.find_matching_gene_name("DAB") is not None

    def test_lizard_inherits_reptilia(self):
        sp = Species.get("Anca")
        assert sp.find_matching_gene_name("UA") is not None


class TestAvesInheritance:
    """Aves sp. provides class I/II genes to all birds."""

    def test_clade_has_genes(self):
        sp = Species.get("Aves sp.")
        assert sp is not None
        for gene in ["UAA", "UBA", "UA", "DAA", "DAB", "DBA", "DBB", "DCA", "DCB", "DRA", "DRB"]:
            assert sp.find_matching_gene_name(gene) is not None, gene

    def test_aves_does_not_have_uca(self):
        """UCA/UDA not on Aves per plan (few/no birds use them)."""
        sp = Species.get("Aves sp.")
        assert sp.find_matching_gene_name("UCA") is None
        assert sp.find_matching_gene_name("UDA") is None

    def test_galliformes_under_aves(self):
        sp = Species.get("Galliformes sp.")
        assert has_ancestor(sp, "Aves sp.")

    def test_aves_to_galliformes_to_chicken_chain(self):
        """Full inheritance chain: Aves → Galliformes → Gallus gallus."""
        sp = Species.get("Gaga")
        assert has_ancestor(sp, "Aves sp.")
        assert has_ancestor(sp, "Galliformes sp.")
        # From Aves
        assert sp.find_matching_gene_name("UAA") is not None
        assert sp.find_matching_gene_name("DRB") is not None
        # From Galliformes
        assert sp.find_matching_gene_name("BF") is not None
        assert sp.find_matching_gene_name("BLB") is not None
        # Own
        assert sp.find_matching_gene_name("BF1") is not None
        assert sp.find_matching_gene_name("YF1") is not None

    def test_ratite_inherits_aves(self):
        sp = Species.get("Stca")
        assert has_ancestor(sp, "Aves sp.")
        assert sp.find_matching_gene_name("UA") is not None
        assert sp.find_matching_gene_name("DAB") is not None
        assert sp.find_matching_gene_name("DRB") is not None

    def test_owl_inherits_aves(self):
        sp = Species.get("Tyal")
        assert has_ancestor(sp, "Aves sp.")
        assert sp.find_matching_gene_name("UA") is not None
        assert sp.find_matching_gene_name("DAB") is not None
        # Owl keeps own numbered loci
        assert sp.find_matching_gene_name("DAB1") is not None
        assert sp.find_matching_gene_name("DAB2") is not None

    def test_penguin_keeps_drb1(self):
        """DRB1 is species-specific, not the same as DRB from Aves."""
        sp = Species.get("Sphu")
        assert sp.find_matching_gene_name("DRB1") is not None
        assert sp.find_matching_gene_name("DRB") is not None  # inherited from Aves

    def test_duck_inherits_aves_plus_own(self):
        sp = Species.get("Anpl")
        assert sp.find_matching_gene_name("UAA") is not None  # from Aves
        assert sp.find_matching_gene_name("DRA") is not None  # from Aves
        assert sp.find_matching_gene_name("UEA") is not None  # own


class TestMarsupialiaInheritance:
    """Marsupialia sp. provides UA, DA, DB, DC, DRA genes to marsupials."""

    def test_clade_has_genes(self):
        sp = Species.get("Marsupialia sp.")
        assert sp is not None
        for gene in ["UA", "DAA", "DAB", "DBA", "DBB", "DCA", "DCB", "DRA"]:
            assert sp.find_matching_gene_name(gene) is not None, gene

    def test_marsupialia_does_not_have_uaa(self):
        """UAA/UBA not on Marsupialia (only Actinopterygii through Aves)."""
        sp = Species.get("Marsupialia sp.")
        assert sp.find_matching_gene_name("UAA") is None
        assert sp.find_matching_gene_name("UBA") is None

    def test_opossum_inherits_marsupial_genes(self):
        sp = Species.get("Modo")
        assert sp.find_matching_gene_name("UA") is not None  # from Marsupialia
        assert sp.find_matching_gene_name("DAA") is not None  # from Marsupialia
        assert sp.find_matching_gene_name("DRA") is not None  # from Marsupialia
        # Own genes
        assert sp.find_matching_gene_name("UA1") is not None
        assert sp.find_matching_gene_name("UB") is not None
        assert sp.find_matching_gene_name("UT3") is not None

    def test_devil_inherits_plus_own(self):
        sp = Species.get("Saha")
        assert sp.find_matching_gene_name("UA") is not None  # from Marsupialia
        assert sp.find_matching_gene_name("DAB") is not None  # from Marsupialia
        assert sp.find_matching_gene_name("UB") is not None  # own
        assert sp.find_matching_gene_name("UK") is not None  # own

    def test_koala_inherits_marsupial(self):
        sp = Species.get("Phci")
        assert sp.find_matching_gene_name("DAA") is not None
        assert sp.find_matching_gene_name("DAB") is not None

    def test_possum_inherits_marsupial(self):
        sp = Species.get("Trvu")
        assert sp.find_matching_gene_name("UA") is not None  # from Marsupialia
        assert sp.find_matching_gene_name("UB") is not None  # own

    def test_marsupial_does_not_get_mammal_drb1(self):
        sp = Species.get("Modo")
        assert sp.find_matching_gene_name("DRB1") is None


class TestCrossTaxonBoundariesWithClades:
    """Ensure clade genes don't leak across major taxon boundaries."""

    def test_mammal_does_not_get_uaa(self):
        """UAA is on Actinopterygii/Amphibia/Reptilia/Aves, not mammals."""
        for prefix in ["HLA", "H2", "BoLA"]:
            sp = Species.get(prefix)
            assert sp.find_matching_gene_name("UAA") is None, prefix

    def test_mammal_does_not_get_uba(self):
        for prefix in ["HLA", "H2", "BoLA"]:
            sp = Species.get(prefix)
            assert sp.find_matching_gene_name("UBA") is None, prefix

    def test_mammal_does_not_get_non_mammalian_dab(self):
        """DAB (non-mammalian class II) should not leak to mammals."""
        for prefix in ["HLA", "H2", "BoLA"]:
            sp = Species.get(prefix)
            assert sp.find_matching_gene_name("DAB") is None, prefix

    def test_fish_does_not_get_dra(self):
        """DRA is on Reptilia/Aves/Marsupialia, not fish."""
        sp = Species.get("Dare")
        assert sp.find_matching_gene_name("DRA") is None

    def test_fish_does_not_get_drb(self):
        sp = Species.get("Dare")
        assert sp.find_matching_gene_name("DRB") is None

    def test_amphibian_does_not_get_dra(self):
        """DRA is not on Amphibia."""
        sp = Species.get("Xela")
        assert sp.find_matching_gene_name("DRA") is None

    def test_shark_inherits_uaa(self):
        """UAA is now on Chondrichthyes (shark literature uses UAA)."""
        sp = Species.get("Gici")
        assert sp.find_matching_gene_name("UAA") is not None

    def test_platypus_stays_at_gnathostomata(self):
        """Platypus is directly under Gnathostomata, no intermediate clade."""
        sp = Species.get("Oran")
        assert not has_ancestor(sp, "Marsupialia sp.")
        assert has_ancestor(sp, "Gnathostomata sp.")


class TestRattusRattusBugFix:
    """Rattus rattus should be under Rattus sp., not Gnathostomata."""

    def test_rattus_rattus_parent(self):
        sp = Species.get("Rara")
        assert sp is not None
        assert has_ancestor(sp, "Rattus sp.")

    def test_rattus_rattus_inherits_rat_genes(self):
        sp = Species.get("Rara")
        assert sp.find_matching_gene_name("Da") is not None



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from typing import Union

from .allele import Allele
from .gene import Gene
from .species import Species

_standard_allele_regex_str = (
    r"([a-zA-Z]+-)?"  # optional species
    r"([a-zA-Z]+\d?|\d{1,2})\*"  # gene, either e.g. "A1" or "88"
    r"(\d{2,3})"  # mandatory first digit field with between 2 and 3 digits
    r"(:\d{2,3})?"  # optional second allele field with up to 3 digits
    r"(:\d\d)?"  # optional third allele field
    r"(:\d\d)?"  # optional fourth allele field
    r"([a-zA-Z])?"  # optional annotation at the end of the allele
)
_standard_allele_regex = re.compile(_standard_allele_regex_str)


def parse_standard_allele_format(
    seq: str, raw_string: Union[str, None] = None, default_species: Union[str, Species, None] = None
):
    """
    Parse alleles which are in a standard format, such as::

        Species-Gene*001:01:01:01

    Between one and four allele fields are allowed and the number of digits
    must be 2 or 3 in the first two fields and exactly two in the last two
    fields. A single character annotation is allowed at the end.

    Parameters
    ----------
    seq : str
        Sequence to parse

    raw_string : str
        Raw string the sequence was derived from

    default_species : str, Species, or None
        If no species is provided in the sequence, should one be assumed?

    Returns
    -------
    Allele or None
    """
    match = _standard_allele_regex.fullmatch(seq)

    if not match:
        return None

    groups = match.groups()

    if len(groups) < 3:
        return None

    species_prefix, gene_name = groups[:2]

    if gene_name is None:
        return None

    if species_prefix is None:
        species = Species.get(default_species)
    elif len(species_prefix) >= 2 and species_prefix[-1] == "-":
        species = Species.get(species_prefix[:-1])
    else:
        return None

    if species is None:
        return None

    gene = Gene.get(species, gene_name)

    if gene is None:
        return None

    allele_fields = []
    for i, raw_allele_field in enumerate(groups[2:-1]):
        if raw_allele_field is None:
            break
        elif i == 0:
            allele_fields.append(raw_allele_field)
        else:
            # skip the initial ':' in all fields after first
            allele_fields.append(raw_allele_field[1:])

    if len(allele_fields) == 0:
        return None

    annotation = groups[-1]

    if annotation:
        annotations = [annotation.upper()]
    else:
        annotations = []

    if raw_string is None:
        raw_string = seq

    return Allele.get_with_gene(gene, allele_fields, annotations=annotations, raw_string=raw_string)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Union

from .common import cache
from .errors import ParseError
from .parser import (
    COLLAPSE_SINGLETON_HAPLOTYPES,
    COLLAPSE_SINGLETON_SEROTYPES,
    DEFAULT_SPECIES_PREFIX,
    GENE_SEPS,
    INFER_CLASS2_PAIRING,
    USE_ALLELE_ALIASES,
    Parser,
)
from .result import Result
from .species import Species


@cache
def cached_parser(
    use_allele_aliases: bool = USE_ALLELE_ALIASES,
    gene_seps=GENE_SEPS,
    collapse_singleton_haplotypes: bool = COLLAPSE_SINGLETON_HAPLOTYPES,
    collapse_singleton_serotypes: bool = COLLAPSE_SINGLETON_SEROTYPES,
    verbose: bool = False,
) -> Parser:
    """
    Get or create a cached Parser instance.

    Construct a Parser instance if this combination of arguments hasn't
    been used before, otherwise retrieve an existing parser from cache.

    Parameters
    ----------
    use_allele_aliases : bool
        Whether to use allele alias mappings for parsing.
    gene_seps : tuple of str
        Separator characters between gene name and allele fields.
    collapse_singleton_haplotypes : bool
        If a Haplotype contains only a single allele, return the allele
        instead of a haplotype.
    collapse_singleton_serotypes : bool
        If a Serotype contains only a single allele, return the allele
        instead of a serotype.
    verbose : bool
        Print intermediate parsing steps for debugging.

    Returns
    -------
    Parser
        A Parser instance configured with the given options.

    Examples
    --------
    >>> parser = cached_parser()
    >>> result = parser.parse("HLA-A*02:01")
    >>> result.gene_name
    'A'
    """
    return Parser(
        use_allele_aliases=use_allele_aliases,
        gene_seps=gene_seps,
        collapse_singleton_haplotypes=collapse_singleton_haplotypes,
        collapse_singleton_serotypes=collapse_singleton_serotypes,
        verbose=verbose,
    )


def _reparse_result_for_species(
    parser: Parser,
    result: Result,
    expected_species: Species,
    infer_class2_pairing: bool,
    required_result_types,
    preferred_result_types,
    only_class1: bool,
    only_class2: bool,
    max_allele_fields,
):
    if type(result) is Species or not hasattr(result, "species"):
        return None

    speciesless_string = result.to_string(include_species=False)
    reparsed = parser.parse(
        f"{expected_species.prefix}-{speciesless_string}",
        default_species=expected_species,
        infer_class2_pairing=infer_class2_pairing,
        raise_on_error=False,
        required_result_types=required_result_types,
        preferred_result_types=preferred_result_types,
        only_class1=only_class1,
        only_class2=only_class2,
        max_allele_fields=max_allele_fields,
    )
    if reparsed is None or type(reparsed) is not type(result):
        return None
    if not hasattr(reparsed, "species") or reparsed.species != expected_species:
        return None
    return reparsed.copy(raw_string=result.raw_string)


def parse(
    raw_string: str,
    default_species=DEFAULT_SPECIES_PREFIX,
    species: Union[str, None] = None,
    use_allele_aliases: bool = USE_ALLELE_ALIASES,
    infer_class2_pairing: bool = INFER_CLASS2_PAIRING,
    collapse_singleton_haplotypes: bool = COLLAPSE_SINGLETON_HAPLOTYPES,
    collapse_singleton_serotypes: bool = COLLAPSE_SINGLETON_SEROTYPES,
    max_allele_fields=None,
    required_result_types=None,
    preferred_result_types=None,
    only_class1: bool = False,
    only_class2: bool = False,
    verbose: bool = False,
    raise_on_error: bool = True,
) -> Optional[Result]:
    """
    Parse MHC alleles into a structured representation.

    Parameters
    ----------
    raw_string : str
       String corresponding to allele, locus, or other MHC-related name

    default_species : str
       By default, parse alleles like "A*02:01" as human but it's possible
       to change this to some other species.

    species : str or None
       Strict species constraint. If provided, the final parse result must have
       this species exactly. Unlike default_species, this does not allow the
       parser to fall back to a different species when the input is explicit
       or ambiguous. If a non-Species result is first parsed at a generic
       ancestor taxon, it may be reparsed for the requested descendant species
       when that conversion is valid.

    use_allele_aliases : bool

    infer_class2_pairing : bool
       If given only the alpha or beta chain of a Class II allele,
       try to infer the most likely pairing from population frequencies.

    collapse_singleton_haplotypes : bool
        If a Haplotype contains only a single allele or Class II allele pair,
        then return the allele instead of a haplotype.

    collapse_singleton_serotypes : bool
        If a Serotype contains only a single allele or Class II allele pair,
        then return the allele instead of a serotype.

    max_allele_fields : int
        If not None, restrict number of allele fields to given value.

    required_result_types : list of type
        Strict filter. Only return results of the given classes.

    preferred_result_types : list of type
        Prefer returning one of these classes when available.
        If none match, return the best non-preferred parse.

    only_class1 : bool
        Only return MHC Class I results

    only_class2 : bool
        Only return MHC Class II results

    verbose : bool
        Print intermediate parsing steps

    raise_on_error : bool
        Raise an exception if string can't be parsed. If False, return None
        instead.

    Returns
    -------
    Result
        A parsed MHC object, which may be an Allele, AlleleWithoutGene, Gene,
        Species, Haplotype, Serotype, Supertype, Pair, Class2Locus, or MhcClass
        depending on the input string. Returns None if parsing fails and
        raise_on_error is False.

    Raises
    ------
    ParseError
        If the string cannot be parsed and raise_on_error is True.

    Examples
    --------
    >>> from mhcgnomes import parse
    >>> allele = parse("HLA-A*02:01")
    >>> allele.gene_name
    'A'
    >>> allele.allele_fields
    ('02', '01')
    >>> allele.to_string()
    'HLA-A*02:01'
    """
    if preferred_result_types is None:
        preferred_result_types = []
    if required_result_types is None:
        required_result_types = []
    parser = cached_parser(
        use_allele_aliases=use_allele_aliases,
        collapse_singleton_haplotypes=collapse_singleton_haplotypes,
        collapse_singleton_serotypes=collapse_singleton_serotypes,
        verbose=verbose,
    )

    # species= is the strict form: the final result must have this species.
    # default_species= is the lenient form: used only when the string
    # doesn't specify a species.
    # They are mutually exclusive.
    if species is not None and default_species != DEFAULT_SPECIES_PREFIX:
        raise ValueError("Cannot specify both 'species' and 'default_species'")

    effective_default = species if species is not None else default_species

    result = parser.parse(
        raw_string,
        default_species=effective_default,
        infer_class2_pairing=infer_class2_pairing,
        raise_on_error=raise_on_error,
        required_result_types=required_result_types,
        preferred_result_types=preferred_result_types,
        only_class1=only_class1,
        only_class2=only_class2,
        max_allele_fields=max_allele_fields,
    )

    # Strict species check: if species= was provided, the final result must
    # have that exact species.
    #
    # Rules:
    # - Species results: exact match only.
    # - Non-Species results with .species: allow ancestor→descendant conversion
    #   only when reparsing under the requested descendant species succeeds.
    # - Descendant→ancestor conversion is never allowed.
    # - Results without .species: always pass.
    if species is not None and result is not None:
        expected = Species.get(species)
        if expected is None:
            # species= was provided but doesn't resolve to a known species.
            # This is always an error — we can't validate against an unknown species.
            if raise_on_error:
                raise ParseError(f"Unknown species '{species}' passed to species= parameter")
            return None
        else:
            if type(result) is Species:
                matches_expected = result == expected
            elif hasattr(result, "species"):
                result_species = result.species
                if result_species == expected:
                    matches_expected = True
                elif result_species.is_ancestor_of(expected):
                    converted = _reparse_result_for_species(
                        parser=parser,
                        result=result,
                        expected_species=expected,
                        infer_class2_pairing=infer_class2_pairing,
                        required_result_types=required_result_types,
                        preferred_result_types=preferred_result_types,
                        only_class1=only_class1,
                        only_class2=only_class2,
                        max_allele_fields=max_allele_fields,
                    )
                    if converted is not None:
                        result = converted
                        matches_expected = True
                    else:
                        matches_expected = False
                else:
                    matches_expected = False
            else:
                matches_expected = True

            if not matches_expected:
                parsed_species = (
                    result.name
                    if type(result) is Species
                    else getattr(result.species, "name", None)
                )
                if raise_on_error:
                    raise ParseError(
                        f"Parsed species '{parsed_species}' does not match "
                        f"expected species '{expected.name}' for '{raw_string}'"
                    )
                return None

    return result



from pathlib import Path

import yaml

from mhcgnomes import Species


def load_underrepresented_taxa_registry():
    registry_path = (
        Path(__file__).parent.parent
        / "mhcgnomes"
        / "data"
        / "underrepresented_taxa_source_registry.yaml"
    )
    with registry_path.open() as fd:
        return yaml.safe_load(fd)


def active_registry_entries():
    return {
        prefix: entry
        for prefix, entry in load_underrepresented_taxa_registry()["taxa"].items()
        if entry.get("curation_status") == "active"
    }


def entry_sources(entry):
    sources = []
    for key in (
        "taxonomy_sources",
        "structured_sources",
        "literature_sources",
        "representative_annotation_sources",
    ):
        sources.extend(entry.get(key, []))
    return sources


def test_registry_captures_mhcseqs_review_prefixes():
    taxa = load_underrepresented_taxa_registry()["taxa"]
    for prefix in [
        "Acar",
        "Acsc",
        "Cyca",
        "Getr",
        "Dare",
        "Orni",
        "Paol",
        "Cyse",
        "Epco",
        "Satr",
        "Chpi",
        "Coja",
        "Fuat",
        "Sthi",
        "Sphu",
        "Spma",
        "Saal",
        "Ritr",
        "Otel",
        "Tyal",
        "Saha",
        "Gaga",
        "Phtr",
        "Phco",
        "Zhom",
    ]:
        assert prefix in taxa, prefix


def test_runtime_ready_registry_entries_match_species_ontology():
    for prefix, entry in active_registry_entries().items():
        species = Species.get(prefix)
        assert species is not None, prefix
        assert species.name == entry["scientific_name"], prefix


def test_active_registry_entries_have_source_backed_prefix_provenance():
    for prefix, entry in active_registry_entries().items():
        sources = entry_sources(entry)
        assert sources, prefix
        assert all(source.startswith("http") for source in sources), prefix


def test_registry_marks_held_back_strings_as_sourced():
    taxa = load_underrepresented_taxa_registry()["taxa"]
    assert taxa["Saal"]["curation_status"] == "active"
    assert taxa["Ritr"]["curation_status"] == "blocked"
    assert taxa["Otel"]["curation_status"] == "blocked"
    assert taxa["Saha"]["curation_status"] == "active"
    assert taxa["Phtr"]["curation_status"] == "blocked"
    assert taxa["Phco"]["curation_status"] == "blocked"
    assert taxa["Zhom"]["curation_status"] == "active"
    assert "Getr-MHC" in " ".join(taxa["Getr"]["ambiguities"])
    assert "Coja-II-13" in " ".join(taxa["Coja"]["ambiguities"] + taxa["Coja"]["blocked_on"])
    assert "prefix collides" in " ".join(taxa["Phco"]["ambiguities"])


def test_registry_tracks_runtime_alias_gap_followups():
    taxa = load_underrepresented_taxa_registry()["taxa"]
    cyca_text = " ".join(map(str, taxa["Cyca"]["observed_structure"] + [taxa["Cyca"]["notes"]]))
    tyal_text = " ".join(map(str, taxa["Tyal"]["observed_structure"] + [taxa["Tyal"]["notes"]]))
    assert "UA1" in cyca_text
    assert "MhcTyal-DAB1" in tyal_text
    assert "YFV" in " ".join(taxa["Gaga"]["ambiguities"] + taxa["Gaga"]["blocked_on"])


def test_registry_tracks_followup_for_runtime_added_coja_ddb1():
    coja = load_underrepresented_taxa_registry()["taxa"]["Coja"]
    assert "DDB1" in coja["canonical_gene_candidates"]
    assert "Coja-DDB1" in " ".join(coja["ambiguities"] + coja["blocked_on"] + [coja["notes"]])



from mhcgnomes import Allele, Gene, Species, parse

from .common import eq_


def test_parse_existing_snake_species_sica():
    expected = Species.get("Sica")
    assert expected is not None
    eq_(parse("Sica"), expected)


def test_parse_existing_snake_gene_sica_daa():
    expected = Gene.get("Sica", "DAA")
    assert expected is not None
    eq_(parse("Sica-DAA"), expected)


def test_parse_existing_snake_allele_sica_daa_01_01():
    expected = Allele.get("Sica", "DAA", "01", "01")
    assert expected is not None
    eq_(parse("Sica-DAA*01:01"), expected)


def test_parse_existing_frog_species_xela():
    expected = Species.get("Xela")
    assert expected is not None
    eq_(parse("Xela"), expected)


def test_parse_existing_frog_gene_xela_uaa():
    expected = Gene.get("Xela", "UAA")
    assert expected is not None
    eq_(parse("Xela-UAA"), expected)


def test_parse_existing_frog_allele_xela_uaa_01_01():
    expected = Allele.get("Xela", "UAA", "01", "01")
    assert expected is not None
    eq_(parse("Xela-UAA*01:01"), expected)


def test_parse_frog_gene_xela_uba():
    expected = Gene.get("Xela", "UBA")
    assert expected is not None
    eq_(parse("Xela-UBA"), expected)


def test_parse_frog_gene_xela_uca():
    expected = Gene.get("Xela", "UCA")
    assert expected is not None
    eq_(parse("Xela-UCA"), expected)


def test_parse_frog_gene_xela_daa():
    expected = Gene.get("Xela", "DAA")
    assert expected is not None
    eq_(parse("Xela-DAA"), expected)


def test_parse_frog_gene_xela_dab():
    expected = Gene.get("Xela", "DAB")
    assert expected is not None
    eq_(parse("Xela-DAB"), expected)


def test_parse_frog_gene_xela_dma():
    expected = Gene.get("Xela", "DMA")
    assert expected is not None
    eq_(parse("Xela-DMA"), expected)


def test_parse_frog_allele_xela_uba_01_01():
    expected = Allele.get("Xela", "UBA", "01", "01")
    assert expected is not None
    eq_(parse("Xela-UBA*01:01"), expected)


def test_parse_frog_allele_xela_dab_01_01():
    expected = Allele.get("Xela", "DAB", "01", "01")
    assert expected is not None
    eq_(parse("Xela-DAB*01:01"), expected)


def test_parse_pilot_frog_species_xetr():
    expected = Species.get("Xetr")
    assert expected is not None
    eq_(parse("Xetr"), expected)
    eq_(Species.get("Xenopus tropicalis"), expected)


def test_parse_frog_gene_xetr_uaa():
    expected = Gene.get("Xetr", "UAA")
    assert expected is not None
    eq_(parse("Xetr-UAA"), expected)


def test_parse_frog_gene_xetr_uba():
    expected = Gene.get("Xetr", "UBA")
    assert expected is not None
    eq_(parse("Xetr-UBA"), expected)


def test_parse_frog_gene_xetr_uca():
    expected = Gene.get("Xetr", "UCA")
    assert expected is not None
    eq_(parse("Xetr-UCA"), expected)


def test_parse_frog_gene_xetr_daa():
    expected = Gene.get("Xetr", "DAA")
    assert expected is not None
    eq_(parse("Xetr-DAA"), expected)


def test_parse_frog_gene_xetr_dab():
    expected = Gene.get("Xetr", "DAB")
    assert expected is not None
    eq_(parse("Xetr-DAB"), expected)


def test_parse_frog_gene_xetr_dma():
    expected = Gene.get("Xetr", "DMA")
    assert expected is not None
    eq_(parse("Xetr-DMA"), expected)


def test_parse_frog_gene_xetr_tap1():
    expected = Gene.get("Xetr", "TAP1")
    assert expected is not None
    eq_(parse("Xetr-TAP1"), expected)


def test_parse_frog_allele_xetr_uaa_01_01():
    expected = Allele.get("Xetr", "UAA", "01", "01")
    assert expected is not None
    eq_(parse("Xetr-UAA*01:01"), expected)


def test_parse_frog_allele_xetr_dab_01_01():
    expected = Allele.get("Xetr", "DAB", "01", "01")
    assert expected is not None
    eq_(parse("Xetr-DAB*01:01"), expected)


# ---------------------------------------------------------------------------
# Other frogs
# ---------------------------------------------------------------------------


def test_parse_omei_tree_frog_species_zhom():
    expected = Species.get("Zhom")
    assert expected is not None
    eq_(parse("Zhom", raise_on_error=True), expected)
    eq_(Species.get("Zhangixalus omeimontis"), expected)


def test_parse_omei_tree_frog_dab():
    expected = Gene.get("Zhom", "DAB")
    assert expected is not None
    eq_(expected.mhc_class, "IIa")
    eq_(parse("Zhom-DAB", raise_on_error=True), expected)


def test_parse_omei_tree_frog_rhom_beta1_alias():
    """Old-genus Rhom-beta1 alias should resolve to DAB."""
    expected = Gene.get("Zhom", "DAB")
    assert expected is not None
    eq_(parse("Zhom-Rhom-beta1", raise_on_error=True), expected)


def test_parse_omei_tree_frog_allele():
    expected = Allele.get("Zhom", "DAB", "01")
    assert expected is not None
    eq_(parse("Zhom-DAB*01", raise_on_error=True), expected)


# ---------------------------------------------------------------------------
# Crocodilians
# ---------------------------------------------------------------------------


def test_parse_crocodilian_species():
    for prefix, latin in [
        ("Crpo", "Crocodylus porosus"),
        ("Meca", "Mecistops cataphractus"),
        ("Oste", "Osteolaemus tetraspis"),
        ("Crni", "Crocodylus niloticus"),
        ("Tosc", "Tomistoma schlegelii"),
    ]:
        expected = Species.get(prefix)
        assert expected is not None, f"Species.get({prefix!r}) returned None"
        eq_(parse(prefix, raise_on_error=True), expected)
        eq_(Species.get(latin), expected)


def test_parse_crpo_expanded_genes():
    """Crpo has class I UA/UB/UC and class II DAA/DAB1/DAB2."""
    for gene_name, mhc_class in [
        ("UA", "I"),
        ("UB", "I"),
        ("UC", "I"),
        ("DAA", "IIa"),
        ("DAB1", "IIa"),
        ("DAB2", "IIa"),
    ]:
        expected = Gene.get("Crpo", gene_name)
        assert expected is not None, f"Gene.get('Crpo', '{gene_name}') returned None"
        eq_(expected.mhc_class, mhc_class)
        eq_(parse(f"Crpo-{gene_name}", raise_on_error=True), expected)


# ---------------------------------------------------------------------------
# Turtles
# ---------------------------------------------------------------------------


def test_parse_turtle_species():
    for prefix, latin in [
        ("CaretCaret", "Caretta caretta"),
        ("Chmy", "Chelonia mydas"),
        ("Deco", "Dermochelys coriacea"),
        ("Leke", "Lepidochelys kempii"),
        ("Chse", "Chelydra serpentina"),
        ("ChrysPicta", "Chrysemys picta"),
        ("Pesi", "Pelodiscus sinensis"),
        ("Tetr", "Terrapene triunguis"),
        ("Gopo", "Gopherus polyphemus"),
        ("Chab", "Chelonoidis abingdonii"),
    ]:
        expected = Species.get(prefix)
        assert expected is not None, f"Species.get({prefix!r}) returned None"
        eq_(parse(prefix, raise_on_error=True), expected)
        eq_(Species.get(latin), expected)


# ---------------------------------------------------------------------------
# Lizards
# ---------------------------------------------------------------------------


def test_parse_pilot_lizard_species_anca():
    expected = Species.get("Anca")
    assert expected is not None
    eq_(parse("Anca"), expected)
    eq_(Species.get("Anolis carolinensis"), expected)


def test_parse_pilot_lizard_species_ansa():
    expected = Species.get("Ansa")
    assert expected is not None
    eq_(parse("Ansa"), expected)
    eq_(Species.get("Anolis sagrei"), expected)


# ---------------------------------------------------------------------------
# Crocodilians
# ---------------------------------------------------------------------------


def test_parse_crocodile_species_crpo():
    expected = Species.get("Crpo")
    assert expected is not None
    eq_(parse("Crpo", raise_on_error=True), expected)
    eq_(Species.get("Crocodylus porosus"), expected)


def test_parse_crocodile_gene_ua():
    expected = Gene.get("Crpo", "UA")
    assert expected is not None
    eq_(expected.mhc_class, "I")
    eq_(parse("Crpo-UA", raise_on_error=True), expected)


def test_parse_crocodile_gene_dab1():
    expected = Gene.get("Crpo", "DAB1")
    assert expected is not None
    eq_(expected.mhc_class, "IIa")
    eq_(parse("Crpo-DAB1", raise_on_error=True), expected)


# ---------------------------------------------------------------------------
# Marsupials
# ---------------------------------------------------------------------------


def test_parse_opossum_species_modo():
    expected = Species.get("Modo")
    assert expected is not None
    eq_(parse("Modo", raise_on_error=True), expected)
    eq_(Species.get("Monodelphis domestica"), expected)


def test_parse_opossum_class1_genes():
    for gene_name in ["UA", "UB", "UC", "UG", "UK", "UM", "UT3", "UT5", "UT7", "UT8"]:
        expected = Gene.get("Modo", gene_name)
        assert expected is not None, f"Gene.get('Modo', '{gene_name}') returned None"
        eq_(expected.mhc_class, "I")
        eq_(parse(f"Modo-{gene_name}", raise_on_error=True), expected)


def test_parse_opossum_class2_dab():
    expected = Gene.get("Modo", "DAB")
    assert expected is not None
    eq_(expected.mhc_class, "IIa")
    eq_(parse("Modo-DAB", raise_on_error=True), expected)


def test_parse_opossum_allele():
    expected = Allele.get("Modo", "UT3", "01")
    assert expected is not None
    eq_(parse("Modo-UT3*01", raise_on_error=True), expected)


def test_parse_devil_species_saha():
    expected = Species.get("Saha")
    assert expected is not None
    eq_(parse("Saha", raise_on_error=True), expected)
    eq_(Species.get("Sarcophilus harrisii"), expected)


def test_parse_devil_class1_genes():
    for gene_name in ["UA", "UB", "UC", "UK", "UM"]:
        expected = Gene.get("Saha", gene_name)
        assert expected is not None, f"Gene.get('Saha', '{gene_name}') returned None"
        eq_(expected.mhc_class, "I")
        eq_(parse(f"Saha-{gene_name}", raise_on_error=True), expected)


def test_parse_devil_class2_dab():
    expected = Gene.get("Saha", "DAB")
    assert expected is not None
    eq_(expected.mhc_class, "IIa")
    eq_(parse("Saha-DAB", raise_on_error=True), expected)



import pytest

from mhcgnomes import Allele, Pair, parse
from mhcgnomes.pair import infer_class2_alpha_chain

from .common import eq_


def test_restrict_num_allele_fields_HLA_DRA_01_01_01_01_DRB1_01_01_01_01():
    result = parse("DRA*01:01:01:01/DRB*01:01:01:01")
    eq_(type(result), Pair)
    eq_(result.alpha.allele_fields, ("01", "01", "01", "01"))
    eq_(result.beta.allele_fields, ("01", "01", "01", "01"))

    result2 = result.restrict_allele_fields(2)
    eq_(
        result2.alpha.allele_fields,
        (
            "01",
            "01",
        ),
    )
    eq_(
        result2.beta.allele_fields,
        (
            "01",
            "01",
        ),
    )


def test_annotation_null_HLA_DRA_01_01_01_01_DRB1_01_01_01_01N():
    result = parse("DRA*01:01:01:01/DRB*01:01:01:01")
    eq_(type(result), Pair)
    assert not result.annotation_null

    result = parse("DRA*01:01:01:01/DRB*01:01:01:01N")
    eq_(type(result), Pair)
    assert result.annotation_null


@pytest.mark.parametrize(
    "annotation, property_name",
    [
        ("C", "annotation_cystosolic"),
        ("S", "annotation_secreted"),
        ("Q", "annotation_questionable"),
        ("L", "annotation_low_expression"),
        ("A", "annotation_aberrant_expression"),
        ("G", "annotation_group"),
        ("Ps", "annotation_pseudogene"),
        ("Sp", "annotation_splice_variant"),
    ],
)
def test_pair_annotation_flags_propagate_from_component_alleles(annotation, property_name):
    alpha = Allele.get("HLA", "DRA", "01", "01", "01", "01")
    beta = Allele.get("HLA", "DRB1", "01", "01", "01", "01", annotation=annotation)
    result = Pair.get(alpha, beta)

    assert result is not None
    assert getattr(result, property_name)


def test_pair_get_rejects_invalid_inputs():
    beta = Allele.get("HLA", "DRB1", "01", "01")

    assert Pair.get(None, beta) is None
    assert Pair.get(beta, None) is None
    assert Pair.get("not-an-allele", beta) is None
    assert Pair.get(beta, "not-an-allele") is None


def test_pair_to_record_and_string_include_gene_and_mhc_class():
    result = parse("HLA-DRA*01:01/DRB1*01:01")

    assert result.to_string() == "HLA-DRA*01:01/DRB1*01:01"
    assert result.gene_name == "DRA/DRB1"
    assert result.to_record() == {
        "gene": "DRA/DRB1",
        "mhc_class": "IIa",
        "is_mutant": False,
        "allele": "HLA-DRA*01:01/DRB1*01:01",
    }


def test_infer_class2_alpha_chain_for_hla_beta_allele_returns_pair():
    beta = Allele.get("HLA", "DRB1", "01", "01")
    result = infer_class2_alpha_chain(beta)

    assert type(result) is Pair
    assert result.alpha == Allele.get("HLA", "DRA", "01", "01")
    assert result.beta == beta


@pytest.mark.parametrize("name", ["HLA-A*02:01", "BoLA-DRB3*001:01"])
def test_infer_class2_alpha_chain_leaves_other_inputs_unchanged(name):
    result = parse(name)

    assert infer_class2_alpha_chain(result) == result



from collections import OrderedDict

import pytest

from mhcgnomes import Allele, Gene, Haplotype, MhcClass, Pair, Serotype, Supertype, parse

ANNOTATION_FLAGS = (
    "annotation_null",
    "annotation_cystosolic",
    "annotation_secreted",
    "annotation_questionable",
    "annotation_low_expression",
    "annotation_aberrant_expression",
    "annotation_group",
    "annotation_pseudogene",
    "annotation_splice_variant",
)


def _annotation_values(result):
    return {flag: getattr(result, flag) for flag in ANNOTATION_FLAGS}


def _expected_annotation_values(*true_flags):
    return {flag: flag in true_flags for flag in ANNOTATION_FLAGS}


GOLDEN_PARSE_CASES = [
    {
        "id": "gene",
        "raw": "HLA-A",
        "expected_type": Gene,
        "expected_to_string": "HLA-A",
        "expected_compact": "A",
        "expected_record": OrderedDict(
            [
                ("species_prefix", "HLA"),
                ("species_name", "Homo sapiens"),
                ("species_latin_name", "Homo sapiens"),
                ("gene", "HLA-A"),
                ("mhc_class", "Ia"),
                ("mutations", ""),
                ("is_mutant", False),
            ]
        ),
        "expected_annotations": _expected_annotation_values(),
    },
    {
        "id": "class_i_allele",
        "raw": "HLA-A*02:01",
        "expected_type": Allele,
        "expected_to_string": "HLA-A*02:01",
        "expected_compact": "A0201",
        "expected_record": OrderedDict(
            [
                ("species_prefix", "HLA"),
                ("species_name", "Homo sapiens"),
                ("species_latin_name", "Homo sapiens"),
                ("gene", "HLA-A"),
                ("mhc_class", "Ia"),
                ("mutations", ""),
                ("is_mutant", False),
                ("allele", "HLA-A*02:01"),
                ("annotations", ()),
            ]
        ),
        "expected_annotations": _expected_annotation_values(),
    },
    {
        "id": "class_ii_pair",
        "raw": "HLA-DRA*01:02/DRB1*03:01",
        "expected_type": Pair,
        "expected_to_string": "HLA-DRA*01:02/DRB1*03:01",
        "expected_compact": "DRA0102-DRB1*0301",
        "expected_record": OrderedDict(
            [
                ("gene", "DRA/DRB1"),
                ("mhc_class", "IIa"),
                ("is_mutant", False),
                ("allele", "HLA-DRA*01:02/DRB1*03:01"),
            ]
        ),
        "expected_annotations": _expected_annotation_values(),
    },
    {
        "id": "mutant_allele",
        "raw": "HLA-A*02:01 N80I mutant",
        "expected_type": Allele,
        "expected_to_string": "HLA-A*02:01 N80I mutant",
        "expected_compact": "A0201 N80I mutant",
        "expected_record": OrderedDict(
            [
                ("species_prefix", "HLA"),
                ("species_name", "Homo sapiens"),
                ("species_latin_name", "Homo sapiens"),
                ("gene", "HLA-A"),
                ("mhc_class", "Ia"),
                ("mutations", "N80I"),
                ("is_mutant", True),
                ("allele", "HLA-A*02:01 N80I mutant"),
                ("annotations", ()),
            ]
        ),
        "expected_annotations": _expected_annotation_values(),
    },
    {
        "id": "pseudogene_allele",
        "raw": "Saoe-G*03:12ps",
        "expected_type": Allele,
        "expected_to_string": "Saoe-G*03:12Ps",
        "expected_compact": "G0312",
        "expected_record": OrderedDict(
            [
                ("species_prefix", "Saoe"),
                ("species_name", "Saguinus oedipus"),
                ("species_latin_name", "Saguinus oedipus"),
                ("gene", "Saoe-G"),
                ("mhc_class", "Ib"),
                ("mutations", ""),
                ("is_mutant", False),
                ("allele", "Saoe-G*03:12Ps"),
                ("annotations", ("Ps",)),
            ]
        ),
        "expected_annotations": _expected_annotation_values("annotation_pseudogene"),
    },
    {
        "id": "serotype",
        "raw": "A2",
        "expected_type": Serotype,
        "expected_to_string": "HLA-A2",
        "expected_compact": "A2",
        "expected_record": OrderedDict(
            [
                ("species_prefix", "HLA"),
                ("species_name", "Homo sapiens"),
                ("species_latin_name", "Homo sapiens"),
                ("serotype", "HLA-A2"),
            ]
        ),
        "expected_annotations": _expected_annotation_values(),
    },
    {
        "id": "haplotype",
        "raw": "H2-k haplotype",
        "expected_type": Haplotype,
        "expected_to_string": "H2-k",
        "expected_compact": "k",
        "expected_record_error": NotImplementedError,
        "expected_annotations": _expected_annotation_values(),
    },
    {
        "id": "supertype",
        "raw": "A2 supertype",
        "expected_type": Supertype,
        "expected_to_string": "HLA A02 supertype",
        "expected_compact": "A02",
        "expected_record": OrderedDict(
            [
                ("species_prefix", "HLA"),
                ("species_name", "Homo sapiens"),
                ("species_latin_name", "Homo sapiens"),
                ("supertype", "A02"),
                ("supertype_string", "HLA A02 supertype"),
                ("representative_allele", "HLA-A*02:01"),
            ]
        ),
        "expected_annotations": _expected_annotation_values(),
    },
    {
        "id": "mhc_class_chain",
        "raw": "HLA class II beta",
        "expected_type": MhcClass,
        "expected_to_string": "human class II beta",
        "expected_compact": "human class II beta",
        "expected_record": OrderedDict(
            [
                ("species_prefix", "HLA"),
                ("species_name", "Homo sapiens"),
                ("species_latin_name", "Homo sapiens"),
                ("mhc_class", "II"),
                ("chain", "beta"),
            ]
        ),
        "expected_annotations": _expected_annotation_values(),
    },
]


@pytest.mark.parametrize(
    "case", GOLDEN_PARSE_CASES, ids=[case["id"] for case in GOLDEN_PARSE_CASES]
)
def test_golden_parser_output_contract(case):
    result = parse(case["raw"])

    assert type(result) is case["expected_type"]
    assert result.to_string() == case["expected_to_string"]
    assert result.compact_string() == case["expected_compact"]
    assert _annotation_values(result) == case["expected_annotations"]

    record_error = case.get("expected_record_error")
    if record_error is not None:
        with pytest.raises(record_error):
            result.to_record()
    else:
        assert result.to_record() == case["expected_record"]


@pytest.mark.parametrize(
    "case", GOLDEN_PARSE_CASES, ids=[case["id"] for case in GOLDEN_PARSE_CASES]
)
def test_cached_parse_identity_and_semantics_are_stable(case):
    first = parse(case["raw"])
    second = parse(case["raw"])

    assert first is second
    assert second == first
    assert hash(second) == hash(first)
    assert second.to_string() == first.to_string()
    assert second.to_dict() == first.to_dict()


@pytest.mark.parametrize(
    "case", GOLDEN_PARSE_CASES, ids=[case["id"] for case in GOLDEN_PARSE_CASES]
)
def test_copy_with_new_raw_string_preserves_semantics(case):
    original = parse(case["raw"])
    copied = original.copy(raw_string=f"copy:{case['raw']}")

    assert type(copied) is type(original)
    assert copied == original
    assert hash(copied) == hash(original)
    assert copied.to_string() == original.to_string()
    assert copied.to_dict() != original.to_dict()
    assert copied.raw_string == f"copy:{case['raw']}"


COPY_CHANGE_CASES = [
    pytest.param("HLA-A", lambda result: result.copy(name="B"), id="gene"),
    pytest.param(
        "HLA-A*02:01",
        lambda result: result.copy(allele_fields=("03", "01")),
        id="class_i_allele",
    ),
    pytest.param(
        "HLA-DRA*01:02/DRB1*03:01",
        lambda result: result.copy(beta=parse("HLA-DRB1*04:01", infer_class2_pairing=False)),
        id="class_ii_pair",
    ),
    pytest.param(
        "HLA-A*02:01 N80I mutant",
        lambda result: result.copy(mutations=()),
        id="mutant_allele",
    ),
    pytest.param(
        "Saoe-G*03:12ps",
        lambda result: result.copy(annotations=()),
        id="pseudogene_allele",
    ),
    pytest.param("A2", lambda result: result.copy(name="A3"), id="serotype"),
    pytest.param("H2-k haplotype", lambda result: result.copy(name="d"), id="haplotype"),
    pytest.param("A2 supertype", lambda result: result.copy(name="A03"), id="supertype"),
    pytest.param(
        "HLA class II beta",
        lambda result: result.copy(chain="alpha"),
        id="mhc_class_chain",
    ),
]


@pytest.mark.parametrize("raw, changer", COPY_CHANGE_CASES)
def test_copy_with_semantic_change_changes_equality(raw, changer):
    original = parse(raw)
    changed = changer(original)

    assert type(changed) is type(original)
    assert changed != original
    assert changed.to_dict() != original.to_dict()


@pytest.mark.parametrize(
    "case", GOLDEN_PARSE_CASES, ids=[case["id"] for case in GOLDEN_PARSE_CASES]
)
def test_serialization_parity_for_representative_results(case):
    original = parse(case["raw"])

    from_dict_result = type(original).from_dict(original.to_dict())
    from_tuple_result = type(original).from_tuple(original.to_tuple())

    assert type(from_dict_result) is type(original)
    assert type(from_tuple_result) is type(original)
    assert from_dict_result == original
    assert from_tuple_result == original
    assert from_dict_result.to_string() == original.to_string()
    assert from_tuple_result.to_string() == original.to_string()



from mhcgnomes import (
    Allele,
    Gene,
    Haplotype,
    Mutation,
    Pair,
    Parser,
    Serotype,
    Species,
    parse,
)
from mhcgnomes.result_sorting import pick_best_result

from .common import eq_


def test_parse_lowercase_hla():
    parser = Parser()
    expected = Species.get("HLA")
    assert expected is not None
    result = parser.parse("hla")
    eq_(result, expected)


def test_parse_lowercase_hla_dra_01_01():
    parser = Parser()
    expected = Allele.get("HLA", "DRA", "01", "01")
    assert expected is not None
    result = parser.parse("hla-dra*01:01")
    eq_(result, expected)


def test_parse_DRB1_01_01():
    parser = Parser()
    expected = Allele.get("HLA", "DRB1", "01", "01")
    result = parser.parse("DRB1*01:01")
    eq_(result, expected)


def test_parse_lowercase_drb1_01_01():
    parser = Parser()
    expected = Allele.get("HLA", "DRB1", "01", "01")
    result = parser.parse("drb1*01:01")
    eq_(result, expected)


def test_parse_multiple_expect_unique_A0201():
    parser = Parser()
    expected_single = Allele.get("HLA", "A", "02", "01")
    assert expected_single is not None
    expected = [expected_single]
    candidates = parser.parse_multiple_candidates("HLA-A*02:01")
    eq_(candidates, list(expected))


def test_parse_hla_lowercase_dra_01_01_drb1_01_01():
    parser = Parser()

    expected_alpha = Allele.get("HLA", "DRA", "01", "01")
    expected_beta = Allele.get("HLA", "DRB1", "01", "01")
    expected_result = Pair.get(expected_alpha, expected_beta)
    result = parser.parse("hla-dra*01:01/drb1*01:01")
    eq_(result, expected_result)


def test_parse_hla_lowercase_dra_01_01_drb1_01_01_no_default_species():
    parser = Parser()
    expected_alpha = Allele.get("HLA", "DRA", "01", "01")
    expected_beta = Allele.get("HLA", "DRB1", "01", "01")
    expected_result = Pair.get(expected_alpha, expected_beta)
    result = parser.parse("hla-dra*01:01/drb1*01:01", default_species=None)
    eq_(result, expected_result)


def test_parse_multiple_candidates_HLA_DRA_01_01_DRB1_01_01_G86Y_mutant():
    parser = Parser()
    results = parser.parse_multiple_candidates("HLA-DRA*01:01/DRB1*01:01 G86Y mutant")
    expected_mutation = Mutation.get(pos=86, aa_original="G", aa_mutant="Y")
    expected_alpha = Allele.get("HLA", "DRA", "01", "01")
    expected_beta = Allele.get("HLA", "DRB1", "01", "01", mutations=(expected_mutation,))
    expected_result = Pair.get(expected_alpha, expected_beta)
    eq_(results, [expected_result])


def test_get_haplotypes_for_any_species_BF19():
    parser = Parser()
    results = parser.get_haplotypes_for_any_species("BF19")
    assert len(results) > 0
    result = results[0]
    eq_(type(result), Haplotype)
    eq_(result.name, "BF19")


def test_parse_multiple_candidates_BF19_class_II():
    parser = Parser()
    results = parser.parse_multiple_candidates("BF19 class II")
    result = results[0]
    eq_(type(result), Haplotype)
    eq_(result.name, "BF19")


def test_parse_haplotype_BF19():
    parser = Parser()
    result = parser.parse_haplotype("BF19")
    eq_(type(result), Haplotype)
    eq_(result.name, "BF19")


def test_parse_multiple_candidates_haplotype_BF19():
    parser = Parser()
    results = parser.parse_multiple_candidates("BF19")
    assert len(results) == 1
    result = results[0]
    eq_(type(result), Haplotype)


def test_parse_multiple_candidates_haplotype_bf19_lowercase():
    parser = Parser()
    results = parser.parse_multiple_candidates("bf19")
    assert len(results) == 1
    result = results[0]
    eq_(type(result), Haplotype)


def test_parse_allele_or_gene_candidates_DRB5_0108N():
    seq = "DRB5_0108N"
    parser = Parser()
    results = parser.parse_allele_or_gene_candidates(
        species=Species.get("HLA"), str_after_species=seq
    )
    assert results is not None
    assert len(results) > 0
    result = results[0]
    eq_(type(result), Allele)


def test_parse_DRB5_0108N():
    seq = "DRB5_0108N"
    parser = Parser()
    result = parser.parse(seq)
    assert result is not None
    eq_(type(result), Allele)


def test_parse_HLA_A2():
    seq = "HLA-A2"
    parser = Parser()
    result = parser.parse(seq)
    assert result is not None
    eq_(type(result), Serotype)


def test_parse_multiple_candidates_A2():
    parser = Parser()
    results = parser.parse_multiple_candidates("A2")
    assert len(results) > 0
    assert any(type(result) is Serotype for result in results)


def test_parse_allele_or_gene_candidates_A2():
    parser = Parser()
    results = parser.parse_allele_or_gene_candidates(
        species=Species.get("HLA"), str_after_species="A2"
    )
    assert len(results) == 0, f"Didn't expect any result but got {results}"


def test_parse_parse_allele_or_gene_candidates_BF2():
    parser = Parser()
    results = parser.parse_allele_or_gene_candidates(
        species=Species.get("chicken"), str_after_species="BF2"
    )
    assert len(results) == 1
    result = results[0]
    assert result is not None
    assert type(result) is Gene, result


def test_parse_parse_BF2():
    parser = Parser()
    result = parser.parse("BF2")
    assert result is not None
    assert type(result) is Gene, result


def test_parse_multiple_candidates_BoLA_DRA_DRB31501():
    parser = Parser()
    results = parser.parse_multiple_candidates("BoLA-DRA-DRB31501")
    assert len(results) > 0
    assert any(type(result) is Pair for result in results)


def test_parse_BoLA_DRA_DRB31501():
    parser = Parser()
    result = parser.parse("BoLA-DRA-DRB31501")
    assert result is not None
    assert type(result) is Pair, f"Expected Pair but got: {result}"


def test_parse_H2K():
    parse_fns = [parse, Parser().parse]
    for fn in parse_fns:
        result = fn("H2K")
        assert result is not None


def test_parse_H2_K_valid_types_Gene():
    parse_fns = [parse, Parser().parse]
    for fn in parse_fns:
        result = fn("H2K", required_result_types=[Gene])
        assert type(result) is Gene


def test_parse_H2_K_valid_types_Haplotype():
    parse_fns = [parse, Parser().parse]
    for fn in parse_fns:
        result = fn("H2K", required_result_types=[Haplotype])
        assert type(result) is Haplotype


def test_parse_preferred_result_types_fallback_to_non_preferred():
    parse_fns = [parse, Parser().parse]
    for fn in parse_fns:
        result = fn(
            "HLA-A*02:01",
            preferred_result_types=[Serotype],
            raise_on_error=False,
        )
        assert result is not None
        assert type(result) is Allele


def test_parse_required_result_types_is_strict():
    parse_fns = [parse, Parser().parse]
    for fn in parse_fns:
        result = fn(
            "HLA-A*02:01",
            required_result_types=[Serotype],
            raise_on_error=False,
        )
        assert result is None


def test_parse_H2_IAb_I67F_R70Q_T71K_mutant():
    s = "H2-IAb I67F, R70Q, T71K mutant"
    parse_fns = [parse, Parser().parse]
    for fn in parse_fns:
        result = fn(s)
        assert result is not None
        assert type(result) in (Pair, Allele), f"Wrong result type: {result}"
        if type(result) is Allele:
            eq_(len(result.mutations), 3)
        elif type(result) is Pair:
            eq_(len(result.beta.mutations), 3)


def test_parse_HLA_DRA_01_01_F54C_mutant_DRB1_01_01():
    s = "HLA-DRA*01:01 F54C mutant/DRB1*01:01"
    parse_fns = [parse, Parser().parse]
    for fn in parse_fns:
        result = fn(s)
        assert result is not None
        assert type(result) is Pair
        eq_(len(result.alpha.mutations), 1)
        eq_(len(result.beta.mutations), 0)


def test_raw_strings_from_parse_class_ii():
    alpha = "DPA1*01:05"
    beta = "DPB1*100:01"
    s = f"{alpha}-{beta}"
    parser = Parser()
    result = parser.parse(s)
    eq_(type(result), Pair)
    eq_(result.raw_string, s)
    eq_(result.alpha.raw_string.lower(), alpha.lower())
    eq_(result.beta.raw_string.lower(), beta.lower())


def test_candidates_from_ambiguous_class2_DPA10105_DPB110001():
    s = "DPA10105-DPB110001"
    parser = Parser()
    results = parser.parse_multiple_candidates(s)
    eq_(len(results), 1)
    eq_(
        results[0],
        Pair.get(Allele.get("HLA", "DPA1", "01", "05"), Allele.get("HLA", "DPB1", "100", "01")),
    )


def test_candidates_from_ambiguous_class2_DPB110001():
    s = "DPB110001"
    parser = Parser()
    results = parser.parse_multiple_candidates(s)
    eq_(len(results), 2)
    assert Allele.get("HLA", "DPB1", "100", "01") in results


def test_candidates_from_ambiguous_mouse_class2_IAb():
    s = "H-2-IAb"
    parser = Parser()
    results = parser.parse_multiple_candidates(s)
    eq_(len(results), 2)
    class2 = Pair.get(Allele.get("H2", "AA", "b"), Allele.get("H2", "AB", "b"))
    assert class2 in results
    gene = Gene.get("H2", "AB")
    assert gene in results
    eq_(pick_best_result(results), class2)


def test_parse_species_SLA_3_YDY01():
    parser = Parser()
    species, str_after_species = parser.parse_species("SLA-3-YDY01")
    assert species.is_pig
    eq_(str_after_species.lower(), "3-ydy01")


def test_parse_species_sla_3_ydy01_lower_case():
    parser = Parser()
    species, str_after_species = parser.parse_species("sla-3-ydy01")
    assert species.is_pig
    eq_(str_after_species.lower(), "3-ydy01")


def test_parse_allele_with_gene_MICA_038():
    parser = Parser()
    allele = parser.parse_allele_with_gene(gene=Gene.get("HLA", "MICA"), str_after_gene="038")
    assert allele is not None
    assert allele.is_class1
    eq_(allele.allele_fields, ("038",))


def test_parse_allele_B_27_215N():
    parser = Parser()
    allele = parser.parse_allele_with_gene(gene=Gene.get("HLA", "B"), str_after_gene="27:215N")
    eq_(allele.allele_fields, ("27", "215"))
    eq_(allele.annotations, ("N",))


def test_parse_B_27_215N():
    parser = Parser(verbose=True)
    allele = parser.parse("B*27:215N")
    eq_(allele.allele_fields, ("27", "215"))
    eq_(allele.annotations, ("N",))



import pytest

from mhcgnomes import Allele, Species, Supertype, parse


def test_supertype_strings_and_record_include_representative_allele():
    result = parse("A2 supertype")

    assert result.to_string() == "HLA A02 supertype"
    assert result.compact_string() == "A02"
    assert result.compact_string(include_species=True) == "HLA-A02"
    assert result.to_record() == {
        "species_prefix": "HLA",
        "species_name": "Homo sapiens",
        "species_latin_name": "Homo sapiens",
        "supertype": "A02",
        "supertype_string": "HLA A02 supertype",
        "representative_allele": "HLA-A*02:01",
    }


def test_supertype_inherited_multiple_allele_helpers():
    allele = Allele.get("HLA", "A", "24", "02", annotation="L")
    result = Supertype(
        species=Species.get("HLA"),
        name="A24",
        alleles=[allele],
        representative=allele,
    )

    assert result.num_alleles == 1
    assert result.gene_name == "A"
    assert result.collapse_if_possible() == allele
    assert result.has_allele
    assert result.is_class1
    assert not result.is_class2
    assert result.annotation_low_expression


def test_supertype_restrict_allele_fields_returns_updated_copy():
    allele = Allele.get("HLA", "A", "02", "01", "01", "01")
    result = Supertype(
        species=Species.get("HLA"),
        name="A02",
        alleles=[allele],
        representative=allele,
    )

    restricted = result.restrict_allele_fields(2)

    assert restricted is not result
    assert restricted.alleles[0].allele_fields == ("02", "01")
    assert restricted.representative == allele


def test_supertype_rejects_mixed_species_alleles():
    with pytest.raises(ValueError):
        Supertype(
            species=Species.get("HLA"),
            name="bad",
            alleles=[
                Allele.get("HLA", "A", "02", "01"),
                Allele.get("BoLA", "NC11", "001", "01"),
            ],
        )



"""
Tests that natural-language phrases from papers containing MHC/HLA keywords
are correctly rejected as parse errors, not accidentally parsed as alleles,
genes, or species.
"""

import pytest

from mhcgnomes import parse

# Phrases that contain MHC keywords but are not allele names.
PAPER_PHRASES = [
    # Genetic phenomena
    "loss of heterozygosity in MHC",
    "MHC haplotype diversity",
    "HLA haploinsufficiency",
    "loss of HLA expression",
    "MHC class I downregulation",
    "MHC class II upregulation",
    "HLA loss of heterozygosity",
    # Biological processes
    "MHC-restricted T cell response",
    "MHC antigen presentation pathway",
    "MHC peptide binding groove",
    "HLA-mediated drug hypersensitivity",
    "MHC-dependent natural killer cell education",
    "cross-presentation by MHC class I",
    # Clinical / disease context
    "MHC-associated autoimmune disease",
    "HLA risk alleles for type 1 diabetes",
    "MHC region on chromosome 6",
    "the MHC locus is highly polymorphic",
    "HLA typing by next-generation sequencing",
    # Experimental methods
    "MHC tetramer staining",
    "HLA-matched donor selection",
    "MHC multimer assay",
    "pan-MHC class I antibody W6/32",
    # Evolution / comparative context
    "MHC evolution in teleost fish",
    "trans-species polymorphism at MHC loci",
    "MHC gene duplication in salmonids",
    "birth-and-death evolution of MHC genes",
    # Phrases with species prefixes embedded in non-allele context
    "Gaga BF locus structure",
    "Dare MHC class I gene family",
    "the Patr MHC region is syntenic with HLA",
    # Short misleading fragments
    # NB: "class I" and "class II" intentionally parse as MhcClass results.
    # "MHC I" intentionally parses as the mouse H2-i haplotype.
    "MHC II",
    "class I and class II",
    "beta-2 microglobulin",
    "antigen processing",
    "peptide loading complex",
    # Journal-style references
    "HLA nomenclature committee report 2024",
    "IPD-MHC database update",
    "IMGT/HLA release 3.51.0",
]


@pytest.mark.parametrize("phrase", PAPER_PHRASES)
def test_paper_phrase_does_not_parse(phrase):
    result = parse(phrase, raise_on_error=False)
    assert result is None, f"Expected None for {phrase!r}, got {result!r}"



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# N = null
# L = low surface expression
# S = soluble / secreted
# C = cytoplasm / not on the surface
# A = aberrant expression
# Q = questionable
# G = group of alleles with identical peptide binding region
# P = same as G?
# Ps = pseudogene
# Sp = splice variant
# For more information see:
# - http://hla.alleles.org/nomenclature/naming.html
# - http://hla.alleles.org/alleles/g_groups.html
# - https://www.ebi.ac.uk/ipd/mhc/group/NHP/page/nomenclature/

valid_functional_annotations = {"N", "L", "S", "C", "A", "Q", "G", "P", "Ps", "Sp"}
_sorted_functional_annotations = sorted(valid_functional_annotations, key=len, reverse=True)


def _has_two_digit_prefix(seq, annot_len):
    if len(seq) < annot_len + 2:
        return False
    return seq[-annot_len - 1].isdigit() and seq[-annot_len - 2].isdigit()


def check_for_workshop_prefix(seq):
    """
    Returns either original sequence and empty list or trims "W"
    from the front of the sequence.
    """
    if len(seq) <= 2:
        return seq, []

    # workshop alleles start with W
    if seq[0].lower() == "w" and seq[1].isdigit() and seq[2].isdigit():
        return seq[1:], ["W"]
    else:
        return seq, []


def peel_suffix_annotations(seq):
    """
    Remove suffix functional annotations from end of sequence,
    return string without annotations and list of parsed annotations.
    """
    if len(seq) < 3:
        return seq, []

    functional_annotations = []
    while True:
        found = False
        for annot in _sorted_functional_annotations:
            n_annot_chars = len(annot)
            if not _has_two_digit_prefix(seq, n_annot_chars):
                continue
            if seq[-n_annot_chars:].lower() == annot.lower():
                functional_annotations = [annot, *functional_annotations]
                seq = seq[:-n_annot_chars]
                found = True
                break
        if not found:
            break
    return seq, functional_annotations


def parse_annotations_from_seq(seq):
    """
    Returns pair of:
        - str of remaining sequence to parse
        - list of annotations
    """
    seq, prefix_annotations = check_for_workshop_prefix(seq)
    seq, suffix_annotations = peel_suffix_annotations(seq)
    return seq, tuple(prefix_annotations), tuple(suffix_annotations)


def parse_annotations_from_allele_fields(allele_fields):
    """
    Returns tuple:
        - tuple of allele fields without any annotations
        - prefix annotations (currently only "W")
        - suffix annotations
    """
    if isinstance(allele_fields, str):
        allele_fields = [field.strip() for field in allele_fields.split(":")]
    allele_fields = list(allele_fields)

    allele_fields[0], prefix_annotations = check_for_workshop_prefix(allele_fields[0])

    allele_fields[-1], suffix_annotations = peel_suffix_annotations(allele_fields[-1])
    return (tuple(allele_fields), tuple(prefix_annotations), tuple(suffix_annotations))



import json
import subprocess
import sys


def run_cli(*args):
    return subprocess.run(
        [sys.executable, "-m", "mhcgnomes", *args],
        capture_output=True,
        text=True,
        check=False,
    )


def run_cli_with_stdin(stdin_text, *args):
    return subprocess.run(
        [sys.executable, "-m", "mhcgnomes", *args],
        input=stdin_text,
        capture_output=True,
        text=True,
        check=False,
    )


def test_cli_help():
    result = run_cli("--help")
    assert result.returncode == 0
    assert "usage:" in result.stdout.lower()
    assert "mhcgnomes" in result.stdout.lower()


def test_cli_table_output_for_valid_allele():
    result = run_cli("HLA-A*02:01")
    assert result.returncode == 0
    output = result.stdout
    assert "input" in output
    assert "type" in output
    assert "normalized" in output
    assert "properties" in output
    assert "HLA-A*02:01" in output
    assert "Allele" in output


def test_cli_tsv_output_multiple_inputs():
    result = run_cli("--format", "tsv", "HLA-A*02:01", "DQ2.5")
    assert result.returncode == 0
    lines = [line for line in result.stdout.strip().splitlines() if line]
    assert len(lines) == 3
    assert lines[0].split("\t") == [
        "input",
        "type",
        "normalized",
        "compact",
        "species",
        "gene",
        "mhc_class",
        "properties",
    ]
    assert "HLA-A*02:01" in lines[1]
    assert "DQ2.5" in lines[2]


def test_cli_strict_mode_fails_on_unparseable_input():
    result = run_cli("--strict", "NOT_A_REAL_ALLELE")
    assert result.returncode == 1
    assert "ParseError" in result.stderr


def test_cli_non_strict_shows_parse_error_row():
    result = run_cli("--format", "tsv", "NOT_A_REAL_ALLELE")
    assert result.returncode == 0
    assert "ParseError" in result.stdout


def test_cli_reads_names_from_stdin_when_no_positional_args_are_given():
    result = run_cli_with_stdin("HLA-A*02:01\nDQ2.5\n", "--format", "json")

    assert result.returncode == 0
    rows = json.loads(result.stdout)
    assert [row["input"] for row in rows] == ["HLA-A*02:01", "DQ2.5"]
    assert rows[0]["normalized"] == "HLA-A*02:01"
    assert rows[1]["type"] == "Pair"


def test_cli_no_header_omits_tsv_header_row():
    result = run_cli("--format", "tsv", "--no-header", "HLA-A*02:01")

    assert result.returncode == 0
    lines = [line for line in result.stdout.strip().splitlines() if line]
    assert len(lines) == 1
    assert not lines[0].startswith("input\t")
    assert lines[0].startswith("HLA-A*02:01\tAllele\t")


def test_cli_errors_when_no_input_is_provided():
    result = run_cli_with_stdin("")

    assert result.returncode == 2
    assert "Provide one or more names" in result.stderr



import pytest

from mhcgnomes import Allele, Gene, parse

from .common import eq_


def test_parse_BoLA_N_04801_with_colon_gene_sep():
    # testing NetMHCpan format for BLA-N*048:01 from
    # https://www.ebi.ac.uk/ipd/mhc/allele/?accession=BoLA03255
    allele = parse("BoLA-N:04801")
    eq_(type(allele), Allele)
    eq_(allele.species.prefix, "BoLA")
    eq_(allele.gene.name, "N")
    eq_(allele.allele_fields, ("048", "01"))


def test_parse_BoLA_N_04801():
    # testing compact format for BLA-N*048:01 from
    # https://www.ebi.ac.uk/ipd/mhc/allele/?accession=BoLA03255
    allele = parse("BoLA-N04801")
    eq_(type(allele), Allele)
    eq_(allele.species.prefix, "BoLA")
    eq_(allele.gene.name, "N")
    eq_(allele.allele_fields, ("048", "01"))


def test_parse_BoLA_N_001001_with_colon_sep():
    allele = parse("BoLA-N:00101")
    eq_(type(allele), Allele)
    eq_(allele.species.prefix, "BoLA")
    eq_(allele.gene.name, "N")
    eq_(allele.allele_fields, ("001", "01"))


def test_bola_nc_does_not_parse_as_n_with_letter_suffix():
    assert parse("BoLA-NC", raise_on_error=False) is None


def test_bola_nc_allele_does_not_parse_as_partial_n_allele():
    assert parse("BoLA-NC*001:01", raise_on_error=False) is None


NOVEL_CATTLE_GENE_EXAMPLES = [
    ("BoLA-DRB4*001:01", "BoLA", "DRB4", "IIa", ("001", "01")),
    ("BoLA-DRB5*001:01", "BoLA", "DRB5", "IIa", ("001", "01")),
    ("BoLA-MIC1*001:01", "BoLA", "MIC1", "Ic", ("001", "01")),
    ("BoLA-MIC2*001:01", "BoLA", "MIC2", "Ic", ("001", "01")),
    ("BoLA-MIC3*001:01", "BoLA", "MIC3", "Ic", ("001", "01")),
    ("BoLA-NC11*001:01", "BoLA", "NC11", "Ib", ("001", "01")),
    ("BoLA-NC12*001:01", "BoLA", "NC12", "Ib", ("001", "01")),
    ("BoLA-NC13*001:01", "BoLA", "NC13", "Ib", ("001", "01")),
    ("Bogr-NC12*001:01", "Bogr", "NC12", "Ib", ("001", "01")),
]


@pytest.mark.parametrize(
    "allele_name,species_prefix,gene_name,mhc_class,_fields",
    NOVEL_CATTLE_GENE_EXAMPLES,
)
def test_cattle_novel_gene_lookup(allele_name, species_prefix, gene_name, mhc_class, _fields):
    gene = Gene.get(species_prefix, gene_name)
    assert gene is not None
    eq_(gene.name, gene_name)
    eq_(gene.mhc_class, mhc_class)


@pytest.mark.parametrize(
    "allele_name,species_prefix,gene_name,mhc_class,fields",
    NOVEL_CATTLE_GENE_EXAMPLES,
)
def test_parse_cattle_novel_gene_alleles(allele_name, species_prefix, gene_name, mhc_class, fields):
    allele = parse(allele_name)
    eq_(type(allele), Allele)
    eq_(allele.species.prefix, species_prefix)
    eq_(allele.gene.name, gene_name)
    eq_(allele.allele_fields, fields)

    if mhc_class == "IIa":
        assert allele.is_class2_beta
    else:
        assert allele.is_class1



import pytest

from mhcgnomes import NormalizingDictionary

from .common import eq_


def test_test_NormalizingDictionary_from_dict_to_dict_identity():
    d = {"aA--": 1, "bb": 2}
    eq_(d, NormalizingDictionary.from_dict(d).to_dict())


def test_NormalizingDictionary_case_invariant():
    d = NormalizingDictionary.from_dict({"aA--": 1, "bb": 2})
    eq_(d["AA--"], 1)
    eq_(d["BB"], 2)


def test_NormalizingDictionary_dash_invariant():
    d = NormalizingDictionary.from_dict({"aA--": 1, "bb": 2})
    eq_(d["AA"], 1)
    eq_(d["BB-"], 2)


def test_NormalizingDictionary_copy_is_independent():
    original = NormalizingDictionary.from_dict({"AA--": 1, "bb": 2})
    copied = original.copy()
    copied["cc"] = 3

    assert "cc" not in original
    eq_(copied["CC"], 3)


def test_NormalizingDictionary_original_key_ambiguity_and_pick_best():
    d = NormalizingDictionary(("AA--", 1), ("aa", 2))

    eq_(d.original_keys("A-A"), {"AA--", "aa"})
    with pytest.raises(ValueError):
        d.original_key("A-A")
    eq_(d.original_key("A-A", pick_best_fn=lambda ks: sorted(ks)[0]), "AA--")


def test_NormalizingDictionary_default_value_fn_populates_missing_key():
    d = NormalizingDictionary(default_value_fn=list)

    value = d["missing"]

    eq_(value, [])
    assert "missing" in d


def test_NormalizingDictionary_map_keys_map_values_and_invert():
    d = NormalizingDictionary(("a", 1), ("b", (1, 2)))

    mapped_values = d.map_values(
        lambda value: tuple(value) if isinstance(value, tuple) else value + 10
    )
    mapped_keys = d.map_keys(lambda key: f"{key}{key}")
    inverted = d.invert()

    eq_(mapped_values["a"], 11)
    eq_(mapped_values["b"], (1, 2))
    eq_(mapped_keys["AA"], 1)
    eq_(inverted[1], {"a", "b"})
    eq_(inverted[2], {"b"})


def test_NormalizingDictionary_key_alignment_prefers_sorted_original_keys():
    d = NormalizingDictionary(("bb", 1), ("AA--", 2), ("aa", 3))

    aligned = list(d.keys_aligned_with_values())
    key_sets = dict(d.key_sets_aligned_with_values())

    eq_(len(d), 2)
    eq_(set(aligned), {"AA--", "bb"})
    eq_(key_sets["AA"], {"AA--", "aa"})



"""
Tests for the 48 batch-added species from mhcseqs unknown species report.
Covers prefix parsing, latin name lookup, gene inheritance, cross-taxon
isolation, and edge cases for non-standard prefixes.
"""

import pytest

from mhcgnomes import Species, parse

from .common import eq_

# All 48 new species with their attested prefixes
BATCH_SPECIES = [
    # Fish
    ("Cosp", "Coregonus sp."),
    ("Trsp", "Tropheus sp."),
    ("KareBico", "Kareius bicoloratus"),
    ("LabeInte", "Labeobarbus intermedius"),
    ("ScopMaxi", "Scophthalmus maximus"),
    ("MonoAlbu", "Monopterus albus"),
    ("AstyMexi", "Astyanax mexicanus"),
    ("SinoGrah", "Sinocyclocheilus grahami"),
    ("SiluAsot", "Silurus asotus"),
    ("OryzDanc", "Oryzias dancena"),
    ("OryzLuzo", "Oryzias luzonensis"),
    ("ScatArgu", "Scatophagus argus"),
    ("RutiRuti", "Rutilus rutilus"),
    ("ParaToxo", "Parachondrostoma toxostoma"),
    ("ClarBatr", "Clarias batrachus"),
    # Snakes
    ("NajaNaja", "Naja naja"),
    ("OphiHann", "Ophiophagus hannah"),
    ("CrotAdam", "Crotalus adamanteus"),
    ("NoteScut", "Notechis scutatus"),
    ("PantGutt", "Pantherophis guttatus"),
    ("PseuText", "Pseudonaja textilis"),
    ("LatiLati", "Laticauda laticaudata"),
    # Lizards
    ("AmblCris", "Amblyrhynchus cristatus"),
    ("EublMacu", "Eublepharis macularius"),
    ("SalvMeri", "Salvator merianae"),
    ("PogoVitt", "Pogona vitticeps"),
    ("PodaMura", "Podarcis muralis"),
    ("LaceAgil", "Lacerta agilis"),
    # Birds
    ("PhasColc", "Phasianus colchicus"),
    ("Iibi", "Tympanuchus cupido"),
    ("CicoBoyc", "Ciconia boyciana"),
    ("AgelPhoe", "Agelaius phoeniceus"),
    ("OceaLeuc", "Oceanodroma leucorhoa"),
    ("EudyMino", "Eudyptula minor"),
    ("ColuLivi", "Columba livia"),
    ("EgreGarz", "Egretta garzetta"),
    ("ChroScop", "Chroicocephalus scopulinus"),
    ("PiprFili", "Pipra filicauda"),
    ("SpheMend", "Spheniscus mendiculus"),
    # Amphibians
    ("RhinMari", "Rhinella marina"),
    ("GeotSera", "Geotrypetes seraphini"),
    ("PeloCult", "Pelobates cultripes"),
    # Turtles
    ("GophEvgo", "Gopherus evgoodei"),
    # Mammals
    ("RhinFerr", "Rhinolophus ferrumequinum"),
    ("MyotLuci", "Myotis lucifugus"),
    ("VombUrsi", "Vombatus ursinus"),
    ("HippArmi", "Hipposideros armiger"),
    ("MyotBran", "Myotis brandtii"),
]


@pytest.mark.parametrize("prefix,latin", BATCH_SPECIES)
def test_batch_species_parse_by_prefix(prefix, latin):
    sp = Species.get(prefix)
    assert sp is not None, f"Species.get({prefix!r}) returned None"
    eq_(sp.name, latin)


@pytest.mark.parametrize("prefix,latin", BATCH_SPECIES)
def test_batch_species_parse_by_latin_name(prefix, latin):
    sp = Species.get(latin)
    assert sp is not None, f"Species.get({latin!r}) returned None"
    eq_(sp.prefix, prefix)


@pytest.mark.parametrize("prefix,latin", BATCH_SPECIES)
def test_batch_species_inherit_root_genes(prefix, latin):
    """Every new species should inherit DMA and TAP1 from Gnathostomata."""
    sp = Species.get(prefix)
    assert sp is not None
    assert sp.find_matching_gene_name("DMA") is not None, f"{latin} missing DMA"
    assert sp.find_matching_gene_name("TAP1") is not None, f"{latin} missing TAP1"


# --- Galliform inheritance ---


def test_pheasant_inherits_bf_from_galliformes():
    result = parse("PhasColc-BF", raise_on_error=True)
    eq_(result.species.name, "Phasianus colchicus")


def test_prairie_chicken_inherits_blb_from_galliformes():
    result = parse("Iibi-BLB", raise_on_error=True)
    eq_(result.species.name, "Tympanuchus cupido")


def test_pheasant_full_latin_name_works():
    result = parse("PhasianusColchicus-BF", raise_on_error=True)
    eq_(result.species.name, "Phasianus colchicus")


# --- Testudines inheritance ---


def test_gopherus_evgoodei_inherits_turtle_genes():
    sp = Species.get("Goev")
    assert sp.find_matching_gene_name("UA") is not None
    assert sp.find_matching_gene_name("DAB") is not None


# --- Cross-taxon leak prevention ---


def test_dove_does_not_get_galliform_bf():
    sp = Species.get("Columba livia")
    assert sp.find_matching_gene_name("BF") is None


def test_cobra_does_not_get_galliform_blb():
    sp = Species.get("Naja naja")
    assert sp.find_matching_gene_name("BLB") is None


def test_bat_does_not_get_fish_uaa():
    sp = Species.get("Myotis lucifugus")
    assert sp.find_matching_gene_name("UAA") is None


def test_toad_inherits_drb1_from_amphibia():
    """DRB1 is now on Amphibia — attested in amphibian MHC literature."""
    sp = Species.get("Rhinella marina")
    assert sp.find_matching_gene_name("DRB1") is not None


# --- species= strict parameter ---


def test_species_strict_dma_with_cobra():
    result = parse("DMA", species="Naja naja", raise_on_error=True)
    eq_(result.species.name, "Naja naja")


def test_species_strict_b2m_with_gecko():
    result = parse("B2M", species="Eublepharis macularius", raise_on_error=True)
    eq_(result.species.name, "Eublepharis macularius")


# --- Non-standard prefix edge cases ---


def test_arbr_prefix_for_roach():
    """Arbr is non-standard (old genus?) but attested in UniProt."""
    eq_(Species.get("Arbr").name, "Rutilus rutilus")


def test_pctn_prefix_for_softnose():
    eq_(Species.get("Pctn").name, "Parachondrostoma toxostoma")


def test_iibi_prefix_for_prairie_chicken():
    eq_(Species.get("Iibi").name, "Tympanuchus cupido")


def test_lasc_prefix_for_gull():
    """Lasc from old genus Larus, now Chroicocephalus."""
    eq_(Species.get("Lasc").name, "Chroicocephalus scopulinus")



from mhcgnomes.token import Token
from mhcgnomes.token_canonical_sequences import (
    ALPHA_CHAIN_TOKEN_SEQ,
    CLASS1_TOKEN_SEQ,
)
from mhcgnomes.tokenize import (
    split_token_sequences,
    tokenize,
)

from .common import eq_


def test_Token():
    t = Token("l-q", raw_string="L-Q")
    eq_(t, t)
    eq_(t, "l-q")


def test_tokenize_complex_uniprot_string_for_H2_Lq():
    s = "MHC class I L-q alpha-chain gene (H2 haplotype), exons 1-3 (Fragment) OS=Mus musculus OX=10090 PE=3 SV=1"
    tokenization_result = tokenize(s)
    eq_(tokenization_result.raw_string, s)
    eq_(
        tokenization_result.tokens,
        (
            CLASS1_TOKEN_SEQ,
            "l-q",
            ALPHA_CHAIN_TOKEN_SEQ,
            "gene",
            "h2",
            "haplotype",
        ),
    )
    eq_(
        tokenization_result.tokens,
        (
            Token(CLASS1_TOKEN_SEQ),
            Token("l-q"),
            Token(ALPHA_CHAIN_TOKEN_SEQ),
            Token("gene"),
            Token("h2"),
            Token("haplotype"),
        ),
    )
    eq_(tokenization_result.tokens[1].raw_string, "L-q")
    assert "OS" in tokenization_result.attributes
    eq_(tokenization_result.attributes["OS"], "Mus musculus")
    assert "OX" in tokenization_result.attributes
    eq_(tokenization_result.attributes["OX"], "10090")
    assert "PE" in tokenization_result.attributes
    eq_(tokenization_result.attributes["PE"], "3")
    assert "SV" in tokenization_result.attributes
    eq_(tokenization_result.attributes["SV"], "1")


def test_tokenize_attributes_only():
    s = "OS=Mus musculus OX=10090 PE=3 SV=1"
    tokenization_result = tokenize(s)
    eq_(tokenization_result.raw_string, s)
    eq_(tokenization_result.tokens, ())
    assert "OS" in tokenization_result.attributes
    eq_(tokenization_result.attributes["OS"], "Mus musculus")
    assert "OX" in tokenization_result.attributes
    eq_(tokenization_result.attributes["OX"], "10090")
    assert "PE" in tokenization_result.attributes
    eq_(tokenization_result.attributes["PE"], "3")
    assert "SV" in tokenization_result.attributes
    eq_(tokenization_result.attributes["SV"], "1")


def test_split_token_sequences():
    s = "Class II DRA0101/DRB0101"
    eq_(split_token_sequences(s), ["Class", "II", "DRA0101", "/", "DRB0101"])


def test_tokenize_BoLA_DRA_DRB31501():
    s = "BoLA-DRA-DRB31501"
    eq_(tokenize(s).tokens, (s.lower(),))



"""
Tests for MHC class + chain parsing (MHCIIB, MHCIIA, mhc2b, mhc1, etc.).

These are region labels meaning "class N [alpha/beta], unknown locus"
and should parse as MhcClass with an optional chain restriction,
NOT as Gene objects.
"""

from mhcgnomes import MhcClass, parse

from .common import eq_

# --- Class II beta forms ---


def test_mhciib_uppercase():
    r = parse("MHCIIB", default_species="Struthio camelus")
    assert isinstance(r, MhcClass)
    eq_(r.mhc_class, "II")
    eq_(r.chain, "beta")


def test_mhciib_lowercase():
    r = parse("mhciib", default_species="Struthio camelus")
    assert isinstance(r, MhcClass)
    eq_(r.chain, "beta")


def test_mhciib_mixed_case():
    r = parse("MhcIIB", default_species="Struthio camelus")
    assert isinstance(r, MhcClass)
    eq_(r.chain, "beta")


def test_mhc_dash_iib():
    r = parse("MHC-IIB", default_species="Struthio camelus")
    assert isinstance(r, MhcClass)
    eq_(r.chain, "beta")


def test_mhc2b():
    r = parse("mhc2b", default_species="Struthio camelus")
    assert isinstance(r, MhcClass)
    eq_(r.mhc_class, "II")
    eq_(r.chain, "beta")


# --- Class II alpha forms ---


def test_mhciia():
    r = parse("MHCIIA", default_species="Struthio camelus")
    assert isinstance(r, MhcClass)
    eq_(r.mhc_class, "II")
    eq_(r.chain, "alpha")


def test_mhc2a():
    r = parse("mhc2a", default_species="Struthio camelus")
    assert isinstance(r, MhcClass)
    eq_(r.chain, "alpha")


# --- Class I (no chain) ---


def test_mhc1():
    r = parse("mhc1", default_species="Struthio camelus")
    assert isinstance(r, MhcClass)
    eq_(r.mhc_class, "I")
    eq_(r.chain, None)


def test_mhci():
    r = parse("MHCI", default_species="Struthio camelus")
    assert isinstance(r, MhcClass)
    eq_(r.mhc_class, "I")
    eq_(r.chain, None)


def test_mhcii_no_chain():
    r = parse("MHCII", default_species="Struthio camelus")
    assert isinstance(r, MhcClass)
    eq_(r.mhc_class, "II")
    eq_(r.chain, None)


# --- Output ---


def test_mhciib_to_string_includes_chain():
    r = parse("MHCIIB", default_species="Struthio camelus")
    eq_(r.to_string(), "ostrich class II beta")
    eq_(r.to_string(include_species=False), "class II beta")


def test_mhciia_to_string():
    r = parse("MHCIIA", default_species="Homo sapiens")
    eq_(r.to_string(), "human class II alpha")


def test_mhc1_to_string_no_chain():
    r = parse("mhc1", default_species="Homo sapiens")
    eq_(r.to_string(), "human class I")


def test_mhciib_to_string_round_trips():
    original = parse("MHCIIB", default_species="Struthio camelus")
    reparsed = parse(original.to_string(), raise_on_error=True)
    assert reparsed == original


def test_hla_class_ii_beta_phrase_parses():
    r = parse("HLA class II beta", raise_on_error=True)
    assert isinstance(r, MhcClass)
    eq_(r.mhc_class, "II")
    eq_(r.chain, "beta")


def test_default_species_class_ii_alpha_phrase_parses():
    r = parse("class II alpha", default_species="Struthio camelus", raise_on_error=True)
    assert isinstance(r, MhcClass)
    eq_(r.species.name, "Struthio camelus")
    eq_(r.mhc_class, "II")
    eq_(r.chain, "alpha")


# --- to_record ---


def test_mhciib_to_record_includes_chain():
    r = parse("MHCIIB", default_species="Struthio camelus")
    record = r.to_record()
    eq_(record["mhc_class"], "II")
    eq_(record["chain"], "beta")


def test_mhci_to_record_no_chain():
    r = parse("MHCI", default_species="Homo sapiens")
    record = r.to_record()
    eq_(record["mhc_class"], "I")
    assert "chain" not in record


# --- species= works with MHCIIB ---


def test_mhciib_with_species_strict():
    r = parse("MHCIIB", species="Struthio camelus")
    assert isinstance(r, MhcClass)
    eq_(r.species.name, "Struthio camelus")


def test_mhciib_with_unknown_species_errors():
    assert parse("MHCIIB", species="Ciconia ciconia", raise_on_error=False) is None


# --- Things that should NOT parse as MHC region labels ---


def test_bare_alpha_does_not_parse_as_mhc_region():
    """'alpha' alone should not become MhcClass."""
    r = parse("alpha", default_species="Homo sapiens", raise_on_error=False)
    assert not isinstance(r, MhcClass) or r is None


def test_bare_beta_does_not_parse_as_mhc_region():
    r = parse("beta", default_species="Homo sapiens", raise_on_error=False)
    assert not isinstance(r, MhcClass) or r is None


def test_mhciii_does_not_parse():
    """MHC class III is not a real adaptive immunity class."""
    r = parse("MHCIII", default_species="Homo sapiens", raise_on_error=False)
    assert r is None


def test_mhc3_does_not_parse():
    r = parse("mhc3", default_species="Homo sapiens", raise_on_error=False)
    assert r is None


def test_mhciic_does_not_parse():
    """MHCIIC — no 'C' chain exists."""
    r = parse("MHCIIC", default_species="Homo sapiens", raise_on_error=False)
    assert r is None


def test_class_iii_beta_phrase_does_not_parse():
    r = parse("HLA class III beta", raise_on_error=False)
    assert r is None



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from typing import Union

from .result_with_mhc_class import ResultWithMhcClass
from .species import Species


@dataclass(eq=False, repr=False, frozen=True, init=False)
class AlleleWithoutGene(ResultWithMhcClass):
    """
    Identifier for molecule whose gene is unknown and no proper allele numbering
    is given for.
    """

    name: str = ""

    def __init__(
        self,
        species: Species,
        name: str,
        mhc_class: Union[str, None] = None,
        raw_string: Union[str, None] = None,
    ):
        ResultWithMhcClass.__init__(
            self, species=species, mhc_class=mhc_class, raw_string=raw_string
        )
        self._set_field(self, "name", name)

    def to_string(self, include_species=True, use_old_species_prefix=False):
        """
        Return allele strings like "BoLA-T2C"
        """
        species_str = self.species.to_string(
            include_species=include_species, use_old_species_prefix=use_old_species_prefix
        )
        if species_str:
            return f"{species_str}-{self.name}"
        return self.name

    def compact_string(self, include_species=False, use_old_species_prefix=False):
        """
        Compact representation, e.g. just "T2C" instead of "BoLA-T2C"
        """
        return self.name

    @classmethod
    def get(cls, species, name, raw_string=None):
        species = Species.get(species)
        if not species:
            return None

        if not name:
            return None

        return cls(species=species, name=name, raw_string=raw_string)



from mhcgnomes import Allele, Gene, Species, parse

from .common import eq_


def test_parse_great_reed_warbler_species_Acar():
    expected = Species.get("Acar")
    assert expected is not None
    eq_(parse("Acar", raise_on_error=True), expected)


def test_parse_great_reed_warbler_UA_gene_and_allele():
    expected_gene = Gene.get("Acar", "UA")
    expected_allele = Allele.get("Acar", "UA", "01", "01")
    assert expected_gene is not None
    assert expected_allele is not None
    eq_(parse("Acar-UA", raise_on_error=True), expected_gene)
    eq_(parse("Acar-UA*01:01", raise_on_error=True), expected_allele)


def test_parse_sedge_warbler_species_Acsc():
    expected = Species.get("Acsc")
    assert expected is not None
    eq_(parse("Acsc", raise_on_error=True), expected)


def test_parse_sedge_warbler_UA_gene_and_allele():
    expected_gene = Gene.get("Acsc", "UA")
    expected_allele = Allele.get("Acsc", "UA", "01", "01")
    assert expected_gene is not None
    assert expected_allele is not None
    eq_(parse("Acsc-UA", raise_on_error=True), expected_gene)
    eq_(parse("Acsc-UA*01:01", raise_on_error=True), expected_allele)


def test_parse_common_yellowthroat_species_Getr():
    expected = Species.get("Getr")
    assert expected is not None
    eq_(parse("Getr", raise_on_error=True), expected)


def test_parse_common_yellowthroat_DAB_gene_and_allele():
    expected_gene = Gene.get("Getr", "DAB")
    expected_allele = Allele.get("Getr", "DAB", "01", "01")
    assert expected_gene is not None
    assert expected_allele is not None
    eq_(parse("Getr-DAB", raise_on_error=True), expected_gene)
    eq_(parse("Getr-DAB*01:01", raise_on_error=True), expected_allele)


def test_common_yellowthroat_does_not_expose_generic_MHC_gene_alias():
    species = Species.get("Getr")
    assert species is not None
    assert species.find_matching_gene_name("MHC") is None


def test_parse_golden_pheasant_species_Chpi():
    expected = Species.get("Chpi")
    assert expected is not None
    eq_(parse("Chpi", raise_on_error=True), expected)


def test_parse_golden_pheasant_class1_genes():
    expected_classes = {"IA1": "Ia", "IA2": "Ia", "IA3": "Ib"}
    for gene_name in ["IA1", "IA2", "IA3"]:
        expected = Gene.get("Chpi", gene_name)
        assert expected is not None
        eq_(expected.mhc_class, expected_classes[gene_name])
        eq_(parse(f"Chpi-{gene_name}", raise_on_error=True), expected)


def test_parse_japanese_quail_species_Coja():
    expected = Species.get("Coja")
    assert expected is not None
    eq_(parse("Coja", raise_on_error=True), expected)


def test_parse_japanese_quail_named_class2_beta_loci():
    examples = [
        ("DAB1", "01"),
        ("DBB1", "01"),
        ("DCB1", "02"),
        ("DDB1", "01"),
        ("DEB1", "02"),
        ("DFB1", "01"),
        ("DGB1", "01"),
    ]
    for gene_name, field in examples:
        expected_gene = Gene.get("Coja", gene_name)
        expected_allele = Allele.get("Coja", gene_name, field)
        assert expected_gene is not None
        assert expected_allele is not None
        eq_(expected_gene.mhc_class, "IIa")
        eq_(parse(f"Coja-{gene_name}", raise_on_error=True), expected_gene)
        eq_(parse(f"Coja-{gene_name}*{field}", raise_on_error=True), expected_allele)


def test_parse_japanese_quail_numbered_class2_aliases():
    """Coja-II-01 through II-07 should map to the canonical loci."""
    mappings = [
        ("II-01", "DAB1"),
        ("II-02", "DBB1"),
        ("II-03", "DCB1"),
        ("II-04", "DDB1"),
        ("II-05", "DEB1"),
        ("II-06", "DFB1"),
        ("II-07", "DGB1"),
    ]
    for alias, canonical in mappings:
        expected = Gene.get("Coja", canonical)
        assert expected is not None
        eq_(parse(f"Coja-{alias}", raise_on_error=True), expected)


def test_parse_japanese_quail_numbered_allele():
    """Coja-II-01*01 should parse as DAB1*01."""
    expected = Allele.get("Coja", "DAB1", "01")
    assert expected is not None
    eq_(parse("Coja-II-01*01", raise_on_error=True), expected)


def test_parse_eurasian_coot_species_Fuat():
    expected = Species.get("Fuat")
    assert expected is not None
    eq_(parse("Fuat", raise_on_error=True), expected)


def test_parse_eurasian_coot_DAB_gene_and_example_allele():
    expected_gene = Gene.get("Fuat", "DAB")
    expected_allele = Allele.get("Fuat", "DAB", "199")
    assert expected_gene is not None
    assert expected_allele is not None
    eq_(expected_gene.mhc_class, "IIa")
    eq_(parse("Fuat-DAB", raise_on_error=True), expected_gene)
    eq_(parse("Fuat-DAB*199", raise_on_error=True), expected_allele)


def test_parse_common_tern_species_Sthi():
    expected = Species.get("Sthi")
    assert expected is not None
    eq_(parse("Sthi", raise_on_error=True), expected)


def test_parse_common_tern_family_level_genes_and_alleles():
    examples = [("UA", ("01", "01"), "I"), ("DAB", ("01", "01"), "IIa")]
    for gene_name, fields, mhc_class in examples:
        expected_gene = Gene.get("Sthi", gene_name)
        expected_allele = Allele.get("Sthi", gene_name, fields)
        assert expected_gene is not None
        assert expected_allele is not None
        eq_(expected_gene.mhc_class, mhc_class)
        eq_(parse(f"Sthi-{gene_name}", raise_on_error=True), expected_gene)
        eq_(parse(f"Sthi-{gene_name}*{':'.join(fields)}", raise_on_error=True), expected_allele)


def test_parse_penguin_species_and_DRB1_family():
    for prefix in ["Sphu", "Spma"]:
        species = Species.get(prefix)
        gene = Gene.get(prefix, "DRB1")
        assert species is not None
        assert gene is not None
        eq_(gene.mhc_class, "IIa")
        eq_(parse(prefix, raise_on_error=True), species)
        eq_(parse(f"{prefix}-DRB1", raise_on_error=True), gene)


def test_parse_barn_owl_family_level_and_embedded_prefix_aliases():
    examples = [
        ("Tyal-UA", Gene.get("Tyal", "UA")),
        # MHCIIB is now parsed as MhcClass(II, beta), not Gene(Tyal, DAB)
        ("Tyal-DRB", Gene.get("Tyal", "DRB")),
        ("Tyal-MhcTyal-UA", Gene.get("Tyal", "UA")),
        ("Tyal-MhcTyal-DAB", Gene.get("Tyal", "DAB")),
        ("Tyal-MhcTyal-DAB1", Gene.get("Tyal", "DAB1")),
        ("Tyal-MhcTyal-DAB2", Gene.get("Tyal", "DAB2")),
    ]
    for raw_string, expected in examples:
        assert expected is not None
        eq_(parse(raw_string, raise_on_error=True), expected)


def test_parse_barn_owl_mhctyal_prefix_alleles():
    examples = [
        ("MhcTyal-UA*01:01", Allele.get("Tyal", "UA", "01", "01")),
        ("MhcTyal-DAB1*01:01", Allele.get("Tyal", "DAB1", "01", "01")),
    ]
    for raw_string, expected in examples:
        assert expected is not None
        eq_(parse(raw_string, raise_on_error=True), expected)


def test_parse_chicken_family_level_aliases_without_breaking_specific_loci():
    expected_pairs = [
        ("Gaga-BF", Gene.get("Gaga", "BF")),
        ("Gaga-B-F", Gene.get("Gaga", "BF")),
        ("Gaga-B-LB", Gene.get("Gaga", "BLB")),
        ("Gaga-BLB", Gene.get("Gaga", "BLB")),
        ("Gaga-B-DMA", Gene.get("Gaga", "DMA")),
        ("Gaga-B-DMB2", Gene.get("Gaga", "DMB2")),
        ("Gaga-B-LB12c", Gene.get("Gaga", "B12c")),
        ("Gaga-BF1", Gene.get("Gaga", "BF1")),
        ("Gaga-BF2", Gene.get("Gaga", "BF2")),
        ("Gaga-BF12", Allele.get("Gaga", "BF", "12")),
        # MHCY / YFV aliases map to YF1
        ("Gaga-YFV", Gene.get("Gaga", "YF1")),
        ("Gaga-MHCY", Gene.get("Gaga", "YF1")),
        ("Gaga-Y15", Gene.get("Gaga", "YF1")),
        # Workshop/serological BF aliases
        ("Gaga-BFw", Gene.get("Gaga", "BF")),
        ("Gaga-BFz", Gene.get("Gaga", "BF")),
        ("Gaga-B-F-S", Gene.get("Gaga", "BF")),
        # MHCY2B class II aliases
        ("Gaga-MHCY2B", Gene.get("Gaga", "BLB")),
        ("Gaga-MHCY2B1", Gene.get("Gaga", "BLB1")),
        ("Gaga-B-DMB1", Gene.get("Gaga", "DMB1")),
    ]
    for raw_string, expected in expected_pairs:
        assert expected is not None
        eq_(parse(raw_string, raise_on_error=True), expected)


def test_parse_chinese_egret_species_Egeu():
    expected = Species.get("Egeu")
    assert expected is not None
    eq_(parse("Egeu", raise_on_error=True), expected)


def test_parse_chinese_egret_class1_genes():
    for gene_name in ["UAA", "UBA"]:
        expected = Gene.get("Egeu", gene_name)
        assert expected is not None
        eq_(expected.mhc_class, "Ia")
        eq_(parse(f"Egeu-{gene_name}", raise_on_error=True), expected)


def test_parse_chinese_egret_class2_genes():
    for gene_name in ["DAB1", "DAB2", "DAB3", "DAB4", "DAB5", "DAB6"]:
        expected = Gene.get("Egeu", gene_name)
        assert expected is not None
        eq_(expected.mhc_class, "IIa")
        eq_(parse(f"Egeu-{gene_name}", raise_on_error=True), expected)


def test_parse_chinese_egret_alleles():
    for gene_name in ["UAA", "DAB1", "DAB3"]:
        expected = Allele.get("Egeu", gene_name, "01", "01")
        assert expected is not None
        eq_(parse(f"Egeu-{gene_name}*01:01", raise_on_error=True), expected)


def test_parse_ratite_species():
    for prefix, latin in [
        ("Stca", "Struthio camelus"),
        ("Drno", "Dromaius novaehollandiae"),
        ("Apow", "Apteryx owenii"),
        ("Apma", "Apteryx mantelli"),
        ("CasuaCasua", "Casuarius casuarius"),
        ("Nope", "Nothoprocta perdicaria"),
    ]:
        expected = Species.get(prefix)
        assert expected is not None, f"Species.get({prefix!r}) returned None"
        eq_(parse(prefix, raise_on_error=True), expected)
        eq_(Species.get(latin), expected)


def test_do_not_parse_ambiguous_or_unreviewed_bird_strings():
    for s in [
        "Gaga-B-LBII",
        "Coja-II-13*01",
        "Coja-II-16*01",
        "Coja-II-17*01",
        "Ritr-DRB1",
        "Otel-DAB",
        "Phtr-UA",
        "Phco-UA",
    ]:
        assert parse(s, raise_on_error=False) is None



from mhcgnomes import (
    Allele,
    Class2Locus,
    Gene,
    Haplotype,
    Pair,
    parse,
)

from .common import eq_


def test_mouse_class1_alleles_H2_Kk():
    H2Kk = Allele.get("H2", "K", "k")

    eq_(parse("H2-Kk"), H2Kk)

    # with a hyphen in "H-2"
    eq_(parse("H-2-Kk"), H2Kk)


def test_mouse_class1_alleles_H2_Db():
    H2Db = Allele.get("H2", "D", "b")

    eq_(parse("H2-Db"), H2Db)

    # with hyphen in "H-2"
    eq_(parse("H-2-Db"), H2Db)


def test_H2_Kd_without_seps():
    eq_(parse("H2Kd"), Allele.get("H2", "K", "d"))


def test_H2_Lq_with_dash_in_species():
    eq_(parse("H-2-Lq"), Allele.get("H2", "L", "q"))


def test_H2_Lq_without_dash_in_species():
    eq_(parse("H2-Lq"), Allele.get("H2", "L", "q"))


def test_mouse_class2_gene():
    # H2-IAb
    gene = Gene.get("H2", "EB2")
    eq_(parse("H2-IEb2"), gene)

    # with hyphen in "H-2"
    eq_(parse("H-2-IEb2"), gene)


def test_parse_H2r():
    haplotype = parse("H2-r")
    assert isinstance(haplotype, Haplotype)
    eq_(haplotype.to_string(), "H2-r")


def test_parse_H2_IE():
    result = parse("H2-IE")
    eq_(type(result), Class2Locus)
    eq_(result.name, "E")


def test_mouse_MR1_weird_uniprot_entry():
    seq = "Major histocompatibility complex class I-related gene protein OS=Mus musculus OX=10090 GN=Mr1 PE=1 SV=2"
    result = parse(seq)
    expected = Gene.get("H2", "MR1")
    eq_(result, expected)


def test_parse_H2_IEd_simplify():
    result = parse("H2-IEd", collapse_singleton_haplotypes=True)
    eq_(type(result), Pair)
    eq_(result.alpha.name, "d")
    eq_(result.beta.name, "d")


def test_parse_H2_IEd_no_simplify():
    result = parse("H2-IEd", collapse_singleton_haplotypes=False)
    eq_(type(result), Haplotype)
    eq_(result.name, "d")
    assert result.locus_restriction is not None



import pytest

from mhcgnomes import (
    MhcClass,
    parse,
)


@pytest.mark.parametrize(
    "s",
    [
        "human class 1",
        "human class i",
        "human class I",
        "hla class 1",
        "hla class i",
        "hla class I",
        "HLA class 1",
        "HLA class i",
        "HLA class I",
    ],
)
def test_human_class_1(s):
    """Test parsing of human class I MHC strings."""
    expected_parsed_result = MhcClass.get("HLA", "I")
    expected_string_repr = "human class I"
    parsed_result = parse(s)
    assert parsed_result == expected_parsed_result, (
        f"Expected {expected_parsed_result} for parsing of '{s}' but got {parsed_result}"
    )
    normalized_str = parsed_result.to_string()
    assert normalized_str == expected_string_repr, (
        f"Expected '{expected_string_repr}' for normalized representation of '{s}' "
        f"but got '{normalized_str}'"
    )
    compact_str = parsed_result.compact_string()
    assert compact_str == expected_string_repr, (
        f"Expected '{expected_string_repr}' for compact representation of '{s}' "
        f"but got '{compact_str}'"
    )


def test_mhc_class_get_rejects_unknown_species_and_invalid_class():
    assert MhcClass.get("NOT_A_SPECIES", "I") is None
    assert MhcClass.get("HLA", "not a class") is None


def test_mhc_class_genes_returns_matching_gene_objects():
    result = MhcClass.get("HLA", "I")

    genes = result.genes()
    gene_names = {gene.name for gene in genes}

    assert "A" in gene_names
    assert "B" in gene_names
    assert "DRA" not in gene_names
    assert all(gene.species_prefix == "HLA" for gene in genes)
    assert all(gene.is_class1 for gene in genes)


def test_mhc_class_to_record_and_strings():
    result = MhcClass.get("HLA", "II")

    assert result is not None
    assert result.is_class2
    assert not result.is_class1
    assert result.to_string() == "human class II"
    assert result.to_string(include_species=False) == "class II"
    assert result.compact_string(include_species=False) == "class II"
    assert result.to_record() == {
        "species_prefix": "HLA",
        "species_name": "Homo sapiens",
        "species_latin_name": "Homo sapiens",
        "mhc_class": "II",
    }



from mhcgnomes import Class2Locus, Species, parse

from .common import eq_


def test_parse_HLA_DR():
    result = parse("HLA-DR")
    eq_(type(result), Class2Locus)
    expected = Class2Locus(species=Species.get("HLA"), name="DR")
    eq_(result, expected)


def test_class2_locus_to_string_without_species_has_no_leading_dash():
    locus = Class2Locus.get("HLA", "DR")
    assert locus is not None
    eq_(locus.to_string(include_species=False), "DR")


def test_class2_locus_compact_string_without_species_has_no_leading_dash():
    locus = Class2Locus.get("HLA", "DQ")
    assert locus is not None
    eq_(locus.compact_string(include_species=False), "DQ")
